import functools
import typing
import random
import time

def retry_on_error(errors: typing.List[typing.Type[BaseException]], max_attempts: int = 10, pause_between_attempts: typing.Union[float, typing.Tuple[float, float]] = 0):
	"""
	Decorator to automatically re-attempt a function when the function encounters one of the given errors.
	
	:param errors: List of exceptions to re-attempt the function for. Matching is done using ``isinstance``.
	:param max_attempts: The maximum number of attempts before the function should truly error.
	:param pause_between_attempts: The time to pause between attempts. If a single number, this is the amount of seconds to pause. If a 2-tuple, this is the range of uniformly-selected random seconds to pause.
	"""
	def wrapper(f):
		@functools.wraps(f)
		def innerwrapper(*args, **kwargs):
			for i in range(max_attempts-1):
				try:
					return f(*args, **kwargs)
				except BaseException as e:
					if not any([ isinstance(e, errortype) for errortype in errors ]):
						raise e
				time.sleep(pause_between_attempts if isinstance(pause_between_attempts, (float, int)) else random.uniform(pause_between_attempts[0], pause_between_attempts[1]))
			return f(*args, **kwargs) # Last attempt
		return innerwrapper
	return wrapper


from esxi_utils.vm.hardware.device import VirtualDevice, VirtualDeviceList
from esxi_utils.util import log, exceptions
import pyVmomi
import typing
import time
import re

class VirtualNICList(VirtualDeviceList):
	"""
	The list of all virtual NICs on a Virtual Machine.
	"""
	def __iter__(self) -> typing.Iterator['VirtualNIC']:
		nics = [ dev for dev in super().__iter__() if isinstance(dev, VirtualNIC) ]
		nics.sort(key=lambda x: int(re.search(r"\d+", x.label).group(0)))
		return iter(nics)

	def __getitem__(self, name_or_index) -> 'VirtualNIC':
		if isinstance(name_or_index, str):
			return self.get(name_or_index)
		else:
			return super().__getitem__(name_or_index)

	def __contains__(self, name: str):
		return self.exists(name)

	def find(self, network: str) -> typing.Union['VirtualNIC', None]:
		"""
		Get the NIC assigned to this VM associated with the given network.

		:param network: The name of the network (portgroup) for the NIC to get.

		:return: A `VirtualNIC` object, or `None` if not found.
		"""
		found = [ nic for nic in self if nic.network == network ]
		if len(found) == 0:
			return None
		if len(found) > 1:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, "VirtualNIC", f"Multiple NICs connected to network {network}")
		return found[0]

	def get(self, network: str) -> 'VirtualNIC':
		"""
		Get the NIC assigned to this VM associated with the given network and raise an exception if not found.

		:param network: The name of the network (portgroup) for the NIC to get.

		:return: A `VirtualNIC` object.
		"""
		nic = self.find(network)
		if nic is None:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, "VirtualNIC", f"Unable to find NIC for network {network}")
		return nic

	def exists(self, network: str) -> bool:
		"""
		Check whether a NIC exists on this virtual machine for the provided network.

		:parma network: The name of the network (portgroup).

		:return: Whether or not a NIC exists connected to this network.
		"""
		return any([nic.network == network for nic in self])

	def add(self, network: str, adapter_type: str = "vmxnet3", pci_slot: typing.Optional[int] = None) -> 'VirtualNIC':
		"""
		Adds a network to the VM.

		:param network: The name of the network (portgroup) to assign to the new nic.
		:param adapter_type: The adapter type to use for the interface
		:param pci_slot: The PCI slot to use for the network interface. For some systems, the interface name is based on the PCI slot 

		:return: The `VirtualNIC` object for the newly added NIC.
		"""
		log.info(f"{str(self)} Adding new virtual NIC with network = {network}, adapter type = {adapter_type}, pci_slot = {pci_slot}")
		assert isinstance(network, str), "network must be a string"
		assert isinstance(adapter_type, str), "adapter_type must be a string"
		assert (isinstance(pci_slot, int) and pci_slot >= 0) or pci_slot is None, "pci_slot must be an integer greater than 0, or None"
		
		# Check that the network exists
		pg = self._vm._client._get_network_object_from_host_system(network)

		ethernet_device_map = {
			"vmxnet": pyVmomi.vim.vm.device.VirtualVmxnet,
			"vmxnet2": pyVmomi.vim.vm.device.VirtualVmxnet2,
			"vmxnet3": pyVmomi.vim.vm.device.VirtualVmxnet3,
			"e1000": pyVmomi.vim.vm.device.VirtualE1000,
			"e1000e": pyVmomi.vim.vm.device.VirtualE1000e,
			"pcnet32": pyVmomi.vim.vm.device.VirtualPCNet32,
			"sriov": pyVmomi.vim.vm.device.VirtualSriovEthernetCard,
		}

		# Get device type
		if adapter_type.lower() not in ethernet_device_map:
			raise exceptions.VirtualMachineInvalidHardwareConfigurationError(self._vm, f"Ethernet device type \"{adapter_type}\" does not exist")
		device = ethernet_device_map[adapter_type]()
		device.wakeOnLanEnabled = True
		device.addressType = 'generated'
		if pci_slot is not None:
			# Check that the PCI slot is not used
			# We need to check this explicitly since an error won't be thrown
			# by default if it is already used
			duplicates = [ n for n in self if str(n.pci) == str(pci_slot) ]
			if len(duplicates) > 0:
				raise exceptions.VirtualMachineInvalidHardwareConfigurationError(self._vm, f"PCI slot {pci_slot} is already in use by NIC {duplicates[0].network}")
			device.slotInfo = pyVmomi.vim.vm.device.VirtualDevice.PciBusSlotInfo(pciSlotNumber=pci_slot)

		# Create device spec
		nic_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		nic_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.add
		nic_spec.device = device

		if isinstance(pg, pyVmomi.vim.dvs.DistributedVirtualPortgroup):
			nic_spec.device.backing = pyVmomi.vim.vm.device.VirtualEthernetCard.DistributedVirtualPortBackingInfo()
			nic_spec.device.backing.port = pyVmomi.vim.DistributedVirtualSwitchPortConnection()
			dpg = self._find_distributed_portgroup(pg._moId)
			if dpg is None:
				raise exceptions.ESXiAPIObjectNotFoundError(pyVmomi.vim.Network, f"distributed network: {network} ... not found")
			nic_spec.device.backing.port.switchUuid = dpg.switchUuid
			nic_spec.device.backing.port.portgroupKey = dpg.portgroupKey
		else: # not distributed
			nic_spec.device.backing = pyVmomi.vim.vm.device.VirtualEthernetCard.NetworkBackingInfo()
			nic_spec.device.backing.useAutoDetect = False
			nic_spec.device.backing.deviceName = network

		nic_spec.device.connectable = pyVmomi.vim.vm.device.VirtualDevice.ConnectInfo()
		nic_spec.device.connectable.startConnected = True
		nic_spec.device.connectable.allowGuestControl = True
		nic_spec.device.connectable.connected = True
		nic_spec.device.connectable.status = 'untried'

		return self._add_device(nic_spec)

	def _find_distributed_portgroup(self, moId_or_portgroupKey: str) -> pyVmomi.vim.dvs.DistributedVirtualPortgroupInfo:
		"""
		Finds the distributed portgroup with the given _moId or portgroupKey by querying the Environment Browser
		for the portgroupKey that matches the NIC's port backing.
		"""
		all_distributed_portgroups = self._vm._client._host_system.parent.environmentBrowser.QueryConfigTarget().distributedVirtualPortgroup
		for dpg in all_distributed_portgroups:
			if dpg.portgroupKey == moId_or_portgroupKey:
				return dpg
		return None

class VirtualNIC(VirtualDevice):
	@property
	def network(self) -> str:
		"""
		The name of the network associated with this NIC.
		"""
		if self.distributed:
			dpg = self._get_distributed_portgroup
			return dpg.portgroupName if dpg is not None else ""
		return self._obj.backing.deviceName

	@network.setter
	def network(self, value: str):
		"""
		Set the network associated with this NIC.
		"""
		log.info(f"{str(self)} Updating network to: \"{value}\"")
		if self.distributed:
			raise exceptions.VirtualMachineInvalidHardwareConfigurationError(self._vm, "Network assignment to a distributed virtual ethernet card is not supported")
		
		self._vm._client._get_network_object_from_host_system(value) # Check that the network to add exists	
		nic_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		nic_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.edit
		nic_spec.device = self._obj
		nic_spec.device.backing.deviceName = value
		
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [nic_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	@property
	def _get_distributed_portgroup(self) -> pyVmomi.vim.dvs.DistributedVirtualPortgroupInfo:
		"""
		Finds the distributed portgroup associated with this nic by querying the Environment Browser
		for the portgroupKey that matches the NIC's port backing.
		"""
		all_distributed_portgroups = self._vm._client._host_system.parent.environmentBrowser.QueryConfigTarget().distributedVirtualPortgroup
		for dpg in all_distributed_portgroups:
			if dpg.portgroupKey == self._obj.backing.port.portgroupKey:
				return dpg
		return None

	@property
	def pci(self) -> int:
		"""
		The PCI slot number associated with this NIC.
		"""
		try:
			return self._obj.slotInfo.pciSlotNumber
		except AttributeError as e:
			if 'pciSlotNumber' in str(e):
				log.critical(f'Unable to determine PCI slot number for network interface: {self} ... Remove the interface and add it again.')
			raise e

	@property
	def mac(self) -> str:
		"""
		The MAC address associated with this NIC.
		"""
		return self._obj.macAddress

	@property
	def distributed(self) -> bool:
		"""
		Whether this virtual ethernet card connects to a distributed virtual switch port or portgroup.
		"""
		return isinstance(self._obj.backing, pyVmomi.vim.vm.device.VirtualEthernetCard.DistributedVirtualPortBackingInfo)

	@property
	def ip(self) -> typing.Union[str, None]:
		"""
		Attempt to detect the VM's IPv4 address on this NIC.
		"""
		# Try to detect with VMware tools first as this is more accurate (and faster)
		if self._vm.tools.running:
			for network in self._vm.tools.networks:
				if network["mac"] == self.mac:
					return network["ips"][0]

		# Fall back to detecting with esxcli
		with self._vm._client.ssh() as conn:
			for port in conn.esxcli(f"network vm port list -w {self._vm.get_world_id()}"):
				if port['MACAddress'] == self.mac:
					ip = port['IPAddress']
					if re.match(r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$", ip) is None or ip == "0.0.0.0":
						continue
					return ip
		return None

	def wait(self, retries: int = 60, delay: int = 2) -> typing.Union[str, None]:
		"""
		Waits for a NIC to become available with an IPv4 address.

		:param retries: How many times to retry detecting an IP on the given network before exiting.
		:param delay: How long to pause between retries in seconds.

		:return: The VM's IPv4 address on the network, or `None` if not found.
		"""
		for _ in range(retries):
			ip = self.ip
			if ip:
				return ip
			time.sleep(delay)
		return None

	def __str__(self):
		return f"<{type(self).__name__}(network='{self.network}') for VM='{self._vm.name}'{' (distributed)' if self.distributed else ''}>"



from esxi_utils.vm.virtualmachine import VirtualMachine
from esxi_utils.util.connect.winrm import WinRMConnection
from esxi_utils.vm.tools.windows import WindowsGuestTools
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.vm.types.ostype import OSType

class WindowsVirtualMachine(VirtualMachine):
	"""
	A Windows virtual machine on an ESXi host. 
	"""
	@property
	def winrm(self) -> typing.Type['WinRMConnection']:
		"""
		A WinRM connection object for establishing a remote connection to this Windows VM.
		"""
		return WinRMConnection

	@property
	def ostype(self) -> 'OSType':
		"""
		The `OSType` for this Virtual Machine.
		"""
		from esxi_utils.vm.types.ostype import OSType
		return OSType.Windows
	
	@property
	def tools(self) -> 'WindowsGuestTools':
		"""
		Get the VMware Tools object for this VM. Includes additional methods for interaction with a Windows operating system.
		"""
		return WindowsGuestTools(self)


# from esxi_utils.vm.types.ostype import OSType
# from esxi_utils.vm.types.cisco import CiscoVirtualMachine
# from esxi_utils.vm.types.linux import LinuxVirtualMachine
# from esxi_utils.vm.types.panos import PaloAltoFirewallVirtualMachine
# from esxi_utils.vm.types.windows import WindowsVirtualMachine



from dateutil.parser import parse as parsetime
from esxi_utils.util.response import Response
from esxi_utils.util import log, exceptions
import subprocess
import datetime
import typing
import random
import ntpath
import winrm
import time
import re


class WinRMConnection:
	"""
	A WinRM connection to a Windows remote host.

	:param ip: The IP of the virtual machine.
	:param username: The username to use when logging into the virtual machine. If `domain` is not specified, this username should take the form `user@domain` (if a domain is available).
	:param password: The password to use when logging into the virtual machine.
	:param domain: The domain (if available) for the user.
	:param transport: Transport type for winrm connection. Defaults to NTLM.
	"""
	def __init__(self, ip: str, username: str, password: str, domain: typing.Optional[str] = None, transport: str = "ntlm"):
		if domain:
			username = f"{username}@{domain}"
		self._ip = ip
		self._username = username
		self._password = password
		self._transport = transport
		self._connection = None

	def __enter__(self) -> 'WinRMConnection':
		self.open()
		return self

	def __exit__(self, *args):
		self.close()

	def open(self):
		"""
		Open a new connection. If an connection is already open, it will be closed before opening a new one.
		"""
		if self._connection:
			self.close()
		
		if self._transport == 'kerberos':
			log.debug('Using kinit to create kerberos ticket')
			try:
				p = subprocess.run(args=['kinit', self._username], input=f'{self._password}\n'.encode(), capture_output=True)
				log.debug(f'kinit stdout: {p.stdout}')
				log.debug(f'Kinit stderr: {p.stderr}')
				if p.returncode != 0:
					raise ChildProcessError(p.stderr.decode())
			except FileNotFoundError:
				raise FileNotFoundError('kinit command does not exist. Make sure to install all necessary kerberos packages')
		
		kwargs = dict()
		kwargs['transport'] = self._transport
		self._connection = winrm.Session(self._ip, auth=(self._username, self._password), **kwargs)

	def close(self):
		"""
		Terminate the network connection to the remote end, if open.

		If no connection is open, this method does nothing.
		"""
		self._connection = None

	def wait(self, retries: int = 10, delay: int = 5, keep_open=False) -> bool:
		"""
		Waits until this connection can be established, and then establishes the connection.

		:param retries: How many times to retry connecting before exiting. Authentication errors will terminate the wait early.
		:param delay: How long to pause between retries in seconds.
		:param keep_open: Whether or not the keep the connection open. If `True`, it is left to the user to close the connection.

		:return: Whether or not the connection could be established
		"""
		from winrm.exceptions import AuthenticationError, BasicAuthDisabledError, InvalidCredentialsError
		fault_tolerance = 2 # When a VM is booting, we may get false errors that resolve with time; allow this many hard failures before exiting
		for i in range(retries):
			log.debug(f"connection attempt {i}")
			if i != 0:
				time.sleep(delay)
			try:
				self.open()
				log.debug(f"open session")
				self.powershell("ls")
				return True
			except (AuthenticationError, BasicAuthDisabledError, InvalidCredentialsError) as e:
				if fault_tolerance == 0:
					raise e
				else:
					fault_tolerance -= 1
					time.sleep(random.randint(20, 40))
			except Exception:
				continue
			finally:
				if not keep_open:
					self.close()
		return False

	def _exec(self, cmd: str, cmdtype: str, assert_status: typing.Optional[int] = None) -> 'Response':
		"""
		Execute a command on the remote end of this connection.

		:param cmd: The command to run.
		:param cmdtype: The type of command to run (either 'powershell' or 'cmd')
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.

		:return: A `Response` object.
		"""
		if self._connection is None:
			raise exceptions.RemoteConnectionNotOpenError(self)
		log.info(f"{str(self)} Running '{cmdtype}' command: {cmd}")
		resp = None
		if cmdtype == "cmd":
			resp = self._connection.run_cmd(cmd)
		elif cmdtype == "powershell":
			resp = self._connection.run_ps(cmd)
		else:
			raise KeyError("Unknown command type: " + cmdtype)
		result = Response(cmd=cmd, stdout=resp.std_out.decode(), stderr=resp.std_err.decode(), status=resp.status_code)
		if assert_status is not None and assert_status != result.status:
			raise exceptions.RemoteConnectionCommandError(self, result)
		return result
		
	def powershell(self, cmd: str, assert_status: typing.Optional[int] = None) -> 'Response':
		"""
		Execute a powershell command on the remote end of this connection.

		:param cmd: The command to run.
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.

		:return: A `Response` object.
		"""
		return self._exec(cmd, cmdtype="powershell", assert_status=assert_status)
		
	def cmd(self, cmd: str, assert_status: typing.Optional[int] = None) -> 'Response':
		"""
		Execute a cmd command on the remote end of this connection.

		:param cmd: The command to run.
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.

		:return: A `Response` object.
		"""
		return self._exec(cmd, cmdtype="cmd", assert_status=assert_status)

	def isfile(self, path: str) -> bool:
		"""
		Checks if the file exists and is a regular file.

		:param path: A path to a file on the remote system.
		
		:return: A boolean whether or not the file exists.
		"""
		return self.powershell(f"(Get-Item \"{path}\") -is [System.IO.FileInfo]").stdout.lower().strip() == "true"

	def isdir(self, path: str) -> bool:
		"""
		Checks if the directory exists.

		:param path: A path to a directory on the remote system.
		
		:return: A boolean whether or not the directory exists.
		"""
		return self.powershell(f"(Get-Item \"{path}\") -is [System.IO.DirectoryInfo]").stdout.lower().strip() == "true"

	def touch(self, path: str):
		"""
		Create an empty file.

		:param path: Path to the new file on the remote system.
		"""
		self.powershell(f"New-Item -Path \"{path}\" -ItemType File", assert_status=0)

	def mkdir(self, path: str, parents: bool = False):
		"""
		Create a directory.

		:param path: Path to the new directory on the remote system.
		:param parents: Whether or not the create parent directories as needed.
		"""
		if not parents:
			# The New-Item command makes directories recursive, so if parents=False we need to check explicitly if the parent exists
			parentdir = ntpath.dirname(path)
			if not self.isdir(parentdir):
				raise exceptions.RemoteNotADirectoryError(self, parentdir)
		self.powershell(f"New-Item -Path \"{path}\" -ItemType Directory", assert_status=0)

	def directory_is_empty(self, path: str) -> bool:
		"""
		Checks whether or not the given directory is empty.

		:param path: Path to the directory on the remote system.

		:return: Boolean whether or not `path` is empty. If the directory does not exist, an error will be thrown.
		"""
		return self.powershell(f"(Get-ChildItem \"{path}\" -Force | Select-Object -First 1 | Measure-Object).Count -eq 0", assert_status=0).stdout.lower().strip() == "true"

	def rm(self, path: str, recursive: bool = False):
		"""
		Remove a file or directory.

		:param path: Path to the directory to remove on the remote system.
		:param recursive: Whether or not the delete directories recursively (required to be True for removing directories).
		"""
		if not recursive and self.isdir(path) and not self.directory_is_empty(path):
			# This check is necessary since trying to remove a non-empty directory without the recurse parameter will cause this command to hang
			raise exceptions.RemoteConnectionError(self, f"Failed to remove {path} as the directory is not empty (use recursive=True)")
		self.powershell(f"Remove-Item -Path \"{path}\" -Force" + (" -Recurse" if recursive else ""), assert_status=0)

	def cp(self, src: str, dst: str, recursive: bool = False):
		"""
		Copy a file or directory.

		:param src: The source for the `cp` command. File globs may be used.
		:param dst: The destination for the `cp` command. If the destination exists, it will be overwritten.
		:param recursive: Copy directories and their contents recursively (required to be True for copying directories).
		"""
		self.powershell(f"Copy-Item -Path \"{src}\" -Destination \"{dst}\" -Force" + (" -Recurse" if recursive else ""), assert_status=0)

	def mv(self, src: str, dst: str):
		"""
		Move a file or directory.

		:param src: The source for the `mv` command. File globs may be used.
		:param dst: The destination for the `mv` command. If the destination exists, it will be overwritten.
		"""
		self.powershell(f"Move-Item -Path \"{src}\" -Destination \"{dst}\" -Force", assert_status=0)

	def ls(self, path: str) -> typing.List[str]:
		"""
		List the files within a directory.

		:param path: Path to the directory on the remote system.

		:return: A list of files in the directory
		"""
		return re.split("\n|\r\n", str(self.powershell(f"(Get-ChildItem -Path \"{path}\" -Force -ErrorAction stop).Name", assert_status=0).stdout))

	def list_installed_packages(self) -> typing.List[str]:
		"""
		Returns a list of installed packages on the VM.

		:return: A list of strings.
		"""
		return re.split("\n|\r\n", str(self.powershell(f"(Get-WmiObject -Class Win32_Product).Name", assert_status=0).stdout))

	def uninstall_package(self, name: str):
		"""
		Uninstall a package by name on the VM.

		:param name: The exact name of the package to uninstall (refer to `list_installed_packages`).
		"""
		self.powershell(f"(Get-WmiObject -Class Win32_Product -Filter \"Name='{name}'\").Uninstall()", assert_status=0)

	def restart(self):
		"""
		Perform a graceful restart of the VM using `Restart-Computer -Force`. This will close the current connection.
		"""
		self.powershell("Restart-Computer -Force")
		self.close()

	def shutdown(self):
		"""
		Attempt a graceful shutdown of the VM using `Stop-Computer -Force`. This will close the current connection.
		"""
		self.powershell("Stop-Computer -Force")
		self.close()

	def time(self):
		"""
		Get the current time set on the remote host.

		:return: A `datetime` object.
		"""
		resp = self.powershell(r"Get-Date -UFormat '%A, %B %d %Y, %R:%S %Z00'", assert_status=0)
		return parsetime(resp.stdout.splitlines()[-1].strip()).astimezone(datetime.timezone.utc)

	@staticmethod
	def parse_PS_objects(str):
		obj_strings = str.split('\r\n\r\n')
		key = None
		value = None
		objs = []
		for obj_string in obj_strings:
			obj = {}
			for line in obj_string.splitlines():
				if ' : ' in line:
					# We have a new key value
					key, value = [s.strip() for s in line.split(' : ', 2)]
					obj[key] = value
				elif key:
					# If we get to a new line, we append the value to the last key
					obj[key] = obj[key]+line.strip()
			objs.append(obj)
		return objs

	def get_AD_group(self, name: str) -> typing.Dict[str, typing.Any]:
		"""
		Retrieves AD group information with group members.

		:param name: The name of the Active Directory group.

		:return: An dict containing information about the Active Directory group.
		"""
		resp = self.powershell(f"Get-ADgroup -Identity {name}", assert_status=0)
		group_data = self.parse_PS_objects(resp.stdout)[0]

		return_obj = dict()
		return_obj['name'] = group_data['Name']
		return_obj['distinguished_name'] = [s.strip() for s in group_data['DistinguishedName'].split(',')]
		return_obj['group_category'] = group_data['GroupCategory']
		return_obj['group_scope'] = group_data['GroupScope']

		resp = self.powershell(f"Get-ADGroupMember -Identity {name}", assert_status=0)
		member_data = self.parse_PS_objects(resp.stdout)

		return_obj['members'] = [ { 'name': memb['name'], 'object_class': memb['objectClass'] } for memb in member_data ]
		return return_obj
		
	def __str__(self):
		return f"<{type(self).__name__} {self._username} at {self._ip}>"

	def __repr__(self):
		return str(self)
		


from esxi_utils import file
from esxi_utils import firewall
from esxi_utils import networking
from esxi_utils import util
from esxi_utils import vm

from esxi_utils.client import ESXiClient
from esxi_utils.datastore import Datastore, DatastoreFile, DatastoreList

__all__ = [
	"ESXiClient",
	"Datastore",
	"DatastoreFile"
]


from esxi_utils.networking.physicalnic import PhysicalNIC, PhysicalNICList
from esxi_utils.networking.portgroup import PortGroup, PortGroupList
from esxi_utils.networking.vmkernelnic import VMKernelNIC, VMKernelNICList
from esxi_utils.networking.vswitch import VSwitch, VSwitchList
from esxi_utils.networking.distributedportgroup import DistributedPortGroup, DistributedPortGroupList
from esxi_utils.networking.distributedvswitch import DistributedVSwitch, DistributedVSwitchList

__all__ = [
    "DistributedPortGroup",
    "DistributedVSwitch",
	"PhysicalNIC",
	"PortGroup",
	"VMKernelNIC",
	"VSwitch"
]


from esxi_utils.vm.hardware.device import VirtualDevice, VirtualDeviceList
from esxi_utils.datastore import DatastoreFile
from esxi_utils.util import log, exceptions
import pyVmomi
import typing
import re

class VirtualFloppyList(VirtualDeviceList):
	"""
	The list of all virtual floppy drives on a Virtual Machine.
	"""
	def __iter__(self) -> typing.Iterator['VirtualFloppy']:
		floppies = [ dev for dev in super().__iter__() if isinstance(dev, VirtualFloppy) ]
		floppies.sort(key=lambda x: int(re.search(r"\d+", x.label).group(0)))
		return iter(floppies)

	def __getitem__(self, index) -> 'VirtualFloppy':
		return self.items[index]

	@property
	def items(self) -> 'VirtualFloppy':
		"""
		A list of all items
		"""
		return list(self)

	def add(self, filepath: 'DatastoreFile') -> 'VirtualFloppy':
		"""
		Add a new floppy drive to the VM's devices.

		:param filepath: A `DatastoreFile` object for a file in the datastore to add to the floppy

		:return: The `VirtualFloppy` object for the newly added floppy drive.
		"""
		log.info(f"{str(self)} Adding new floppy drive with file: {filepath}")
		assert isinstance(filepath, DatastoreFile), "filepath must be a DatastoreFile object"

		if len(list(self)) >= 2:
			raise exceptions.VirtualMachineInvalidHardwareConfigurationError(self._vm, "Maximum number of floppy drives (2) reached")

		controllers = VirtualDeviceList(self._vm).find_type("VirtualSIOController")
		if len(controllers) == 0:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, "SIO Controller", f"No controllers present to add floppy drive")
		controller = controllers[0]

		floppy_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		floppy_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.add
		floppy_spec.device = pyVmomi.vim.vm.device.VirtualFloppy()
		floppy_spec.device.controllerKey = controller.key

		floppy_spec.device.backing = pyVmomi.vim.vm.device.VirtualFloppy.ImageBackingInfo()
		floppy_spec.device.backing.fileName = filepath.path

		floppy_spec.device.connectable = pyVmomi.vim.vm.device.VirtualDevice.ConnectInfo()
		floppy_spec.device.connectable.allowGuestControl = True
		floppy_spec.device.connectable.connected = False
		floppy_spec.device.connectable.startConnected = False
		return self._add_device(floppy_spec)


class VirtualFloppy(VirtualDevice):
	@property
	def file(self) -> 'DatastoreFile':
		"""
		The file attached to this floppy as a `DatastoreFile` object.
		"""
		return self._get_backing_file()

	@file.setter
	def file(self, filepath: 'DatastoreFile'):
		"""
		Set the file for this floppy. A floppy cannot be empty, and thus this can only change an attached file for another.

		:param filepath: A `DatastoreFile` object for a file in the datastore to add to the floppy.
		"""
		log.info(f"{str(self)} Updating file to: {filepath}")
		assert isinstance(filepath, DatastoreFile), "filepath must be a DatastoreFile object"

		floppy_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		floppy_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.edit
		floppy_spec.device = self._obj
		floppy_spec.device.backing = pyVmomi.vim.vm.device.VirtualFloppy.ImageBackingInfo()
		floppy_spec.device.backing.fileName = filepath.path
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [floppy_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	def __str__(self):
		return f"<{type(self).__name__}(file='{self.file}') for VM='{self._vm.name}'>"


import typing
import enum

if typing.TYPE_CHECKING:
	from esxi_utils.vm.virtualmachine import VirtualMachine

class OSType(enum.Enum):
	#: OS Unknown
	Unknown = "unknown"

	#: Windows OS
	Windows = "windows"

	#: Linux OS
	Linux = "linux"

	#: Cisco-based OS
	Cisco = "cisco"

	#: Panorama-based OS
	PanOs = "panos"

	@staticmethod
	def map(ostype: typing.Optional['OSType']) -> 'VirtualMachine':
		"""
		Map an OSType value to the appropriate virtual machine class.

		:return: A VirtualMachine type
		"""
		from esxi_utils.vm.virtualmachine import VirtualMachine
		from esxi_utils.vm.types.cisco import CiscoVirtualMachine
		from esxi_utils.vm.types.linux import LinuxVirtualMachine
		from esxi_utils.vm.types.panos import PaloAltoFirewallVirtualMachine
		from esxi_utils.vm.types.windows import WindowsVirtualMachine

		typemap = {
			OSType.Windows: WindowsVirtualMachine,
			OSType.Linux: LinuxVirtualMachine,
			OSType.Cisco: CiscoVirtualMachine,
			OSType.PanOs: PaloAltoFirewallVirtualMachine
		}
		return typemap.get(ostype, VirtualMachine)

	@staticmethod
	def detect(guest_id: str) -> 'OSType':
		"""
		Attempt to detect the OSType from the given guest ID string.

		:param guest_id: The guest ID string.

		:return: The detected OSType.
		"""
		os = OSType.Unknown
		if guest_id and any([ keyword in guest_id.lower() for keyword in [ "ubuntu", "debian", "centos", "rhel", "linux" ] ]):
			os = OSType.Linux
		elif guest_id and any([ keyword in guest_id.lower() for keyword in [ "windows" ] ]):
			os = OSType.Windows
		return os


from esxi_utils.util import log, exceptions
from vncdotool import api as vncdotool_api
import random
import typing
import time
import os

if typing.TYPE_CHECKING:
	from esxi_utils.vm.virtualmachine import VirtualMachine

class VNCHandler:
	"""
	Handler class for using VNC on a VirtualMachine.

	:param vm: A `VirtualMachine` object
	"""
	def __init__(self, vm: 'VirtualMachine'):
		self._vm = vm
		self._client = vm._client

	def disable(self):
		"""
		Disables VNC.
		"""
		log.debug(f"{str(self)} Disabling...")
		self._vm.vmx = { key: value for key, value in self._vm.vmx.items() if not key.lower().startswith("remotedisplay.vnc.") }

	def enable(self, port: typing.Optional[int] = None) -> int:
		"""
		Enables VNC. Note that the ESXi server must be configured to allow ports 5900-6000 for VNC to work.

		:param port: The VNC port to use. If None, the VNC port will be auto-detected.

		:return: The VNC port assigned to this VM.
		"""
		if port is None:
			ports = self._client.firewall.find_vnc_port()
			if len(ports) == 0:
				raise exceptions.FirewallError(self, "Failed to determine a valid VNC port, check your connection and firewall rules to ensure incoming ports 5900-6000 are allowed")
			port = ports[random.randint(0, len(ports)-1)] # Choose a random VNC port from the list of valid ports
		log.debug(f"{str(self)} Enabling on port {port}...")
		vmx = self._vm.vmx
		vmx["remotedisplay.vnc.enabled"] = "TRUE"
		vmx["remotedisplay.vnc.port"] = str(port)
		self._vm.vmx = vmx
		return port

	def url(self) -> typing.Union[str, None]:
		"""
		Returns the VNC URL to this VM. VNC must be enabled on the VM.

		:return: A URL to use for connecting over VNC, or None if VNC is not enabled.
		"""
		vmx = self._vm.vmx
		if "remotedisplay.vnc.port" not in vmx:
			return None
		port = vmx["remotedisplay.vnc.port"]
		return f"{self._client.hostname}::{port}"

	def press_key(self, key: str, timeout: int = 30):
		"""
		Uses VNC to simulate a key press on the remote VM. Raises an exception is VNC is not enabled.

		The key is sent as if a user was writing over a UI. It is up to the user to ensure that the VM
		is in a state to properly receive the key press (i.e. the console is open if trying to write to
		a console).

		:param key: The key to send over VNC.
		:param timeout: The maximum number of seconds to wait for the operation to complete before raising a `TimeoutError`.
		"""
		log.debug(f"{str(self)} Simulating key press: {key}")
		vnc_url = self.url()
		if vnc_url is None:
			raise exceptions.VNCNotEnabledError(self._vm)
		with vncdotool_api.connect(vnc_url) as client:
			client.timeout = timeout
			client.keyPress(key)

	def write(self, text: str, enter: bool = False, timeout: int = 30):
		"""
		Uses VNC to simulate writing text on the remote VM. Raises an exception is VNC is not enabled.

		The keys are sent as if a user was writing over a UI. It is up to the user to ensure that the VM
		is in a state to properly receive the key presses (i.e. the console is open if trying to write to
		a console). 

		:param text: The text to write over VNC.
		:param enter: Boolean value whether to send the `ENTER` key after writing the text.
		:param timeout: The maximum number of seconds to wait for the operation to complete before raising a `TimeoutError`.
		"""
		log.debug(f"{str(self)} Simulating write: {text}")
		vnc_url = self.url()
		if vnc_url is None:
			raise exceptions.VNCNotEnabledError(self._vm)
		with vncdotool_api.connect(vnc_url) as client:
			client.timeout = timeout
			for char in text:
				if char.isupper() or char in r"~!@#$%^&*()_+{}|:\"<>?":
					# Requires shift key to be pressed
					client.keyDown('shift')
					client.keyPress(char)
					client.keyUp('shift')
				else:
					client.keyPress(char)
			if enter:
				client.keyPress('enter')
	
	def capture_screen(self, filename: str, timeout: float = 30.0):
		"""
		Capture a screenshot of the VM's screen.

		:param filename: 
			Path to a file or directory where the screenshot should be captured.
			If a directory, the screenshot will be saved as `capture.png` in that directory.
		:param timeout: 
			The maximum number of seconds to wait for the operation to complete before raising a `TimeoutError`.

		:return: The path to the saved screenshot.
		"""
		assert isinstance(filename, str) and len(filename) > 0, "filename not specified"
		vnc_url = self.url()
		if vnc_url is None:
			raise exceptions.VNCNotEnabledError(self._vm)
		filename = os.path.abspath(filename)
		if os.path.isdir(filename):
			filename = os.path.join(filename, "capture.png")
		if not filename.lower().endswith(".png"):
			filename += ".png"
		if not os.path.isdir(os.path.dirname(filename)):
			raise NotADirectoryError(f"{os.path.dirname(filename)} is not an existing directory")
		with vncdotool_api.connect(vnc_url) as client:
			client.timeout = timeout
			client.refreshScreen()
			client.captureScreen(filename)
		return filename

	def stream_screen(self, dirname: str, timeout: float = 5, continue_on_timeout: bool = True, pause: float = 0.0, continuefunc: typing.Optional[typing.Callable[[], bool]] = None) -> typing.Generator[str, None, None]:
		"""
		Continuously capture screenshots of the VM's screen in a loop. Use a ``break`` statement, ``.close()``, or the ``continuefunc`` parameter to exit the stream.

		Example usage (loop)::

			i = 0
			for screenshot_png in vm.vnc.stream_screen("screenshots/"):
				# screenshot_png = path to the taken screenshot
				# Do something with the screenshot
				i += 1
				if i == 10:
					break
					
		Example usage (next)::

			screenshot_generator = vm.vnc.stream_screen("screenshots/")
			screenshot_1_png = next(screenshot_generator)
			screenshot_2_png = next(screenshot_generator)
			screenshot_3_png = next(screenshot_generator)
			screenshot_generator.close()

		Example usage (callback)::

			i = 0
			for screenshot_png in vm.vnc.stream_screen("screenshots/", continuefunc=lambda: i < 10):
				# screenshot_png = path to the taken screenshot
				# Do something with the screenshot
				i += 1

		:param dirname: The directory in which to save screenshots. Screenshots will be PNG files named with the time they are taken (nanoseconds since epoch).
		:param timeout: The maximum number of seconds to wait per screenshot.
		:param continue_on_timeout: Whether to continue taking screenshots if a timeout is reached (if ``False``, a ``TimeoutError`` is raised instead).
		:param pause: Number of seconds to pause between each screenshot.
		:param continuefunc: An optional function that is called prior to taking each screenshot. If this function returns ``False``, no more screenshots will be taken and this generator will exit.

		:yields: The paths to each screenshot taken.
		"""
		assert isinstance(dirname, str) and len(dirname) > 0, "dirname not specified"
		assert isinstance(timeout, (int, float)) and timeout > 0, "timeout is not a number greater than 0"
		assert isinstance(continue_on_timeout, bool), "continue_on_timeout is not boolean"
		assert isinstance(pause, (int, float)) and pause >= 0, "pause is not a number greater or equal to 0"
		assert continuefunc is None or callable(continuefunc), "continuefunc is not None or a callable"

		vnc_url = self.url()
		if vnc_url is None:
			raise exceptions.VNCNotEnabledError(self._vm)
		
		if not os.path.isdir(dirname):
			raise NotADirectoryError(f"{dirname} is not a directory")
		
		client = vncdotool_api.connect(vnc_url)
		try:
			client.timeout = timeout
			while continuefunc is None or continuefunc() != False:
				try:
					path = os.path.join(dirname, str(time.time_ns()) + ".png")
					client.refreshScreen()
					client.captureScreen(path)
					yield path
				except TimeoutError as e:
					if not continue_on_timeout:
						raise e
				if pause:
					time.sleep(pause)
		finally:
			client.disconnect()

	def expect_screen(self, filename: str, maxrms: int = 0, timeout: int = 30):
		"""
		Wait until the display matches a target image. If this fails to get a match on the display in time, a `TimeoutError` will be raised.

		:param filename: An image file to read and compare against.
		:param maxrms: The maximum root mean square between histograms of the screen and target image.
		:param timeout: The maximum number of seconds to wait for the operation to complete before raising a `TimeoutError`.
		"""
		assert isinstance(filename, str) and len(filename) > 0, "filename not specified"
		vnc_url = self.url()
		if vnc_url is None:
			raise exceptions.VNCNotEnabledError(self._vm)
		filename = os.path.abspath(filename)
		if not os.path.isfile(filename):
			raise FileNotFoundError(f"{filename} is not a file")
		if not filename.lower().endswith(".png"):
			raise ValueError(f"{filename} is not a png file")
		with vncdotool_api.connect(vnc_url) as client:
			client.timeout = timeout
			client.refreshScreen()
			client.expectScreen(filename, maxrms=maxrms)
			
	def __str__(self):
		return f"<{type(self).__name__} for {self._vm.name}>"
	
	def __repr__(self):
		return str(self)
	
	def get_mks_ticket(self, webmks: bool = True):
		"""
		Ask the referenced VM to provide an authentication ticket.
		The requesting user must have the 'VirtualMachine.Interact.ConsoleInteract' privilege.

		:param webmks:
			If True: requests a 'webmks' ticket
			If False: requests a 'mks' ticket
			Both options will return a 'vim.VirtualMachine.Ticket' (read below)
		
		Returns a 'vim.VirtualMachine.Ticket' object for the mks protocol.
		Possible fields in the return object appear to be:
		- dynamicType <unset>
		- dynamicProperty (object)
		- ticket (string)
		- cfgFile (string)
		- host <unset>
		- sslThumbprint (string)
		- url <unset>

		Where <unset> values were observed on a non-vCenter system.
		"""
		ticket_type = 'webmks'
		if not webmks:
			ticket_type = 'mks'
		return self._vm._vim_vm.AcquireTicket(ticket_type)



from esxi_utils.util import log, exceptions, usbkeycodes
import pyVmomi
import typing


if typing.TYPE_CHECKING:
    from esxi_utils.vm.virtualmachine import VirtualMachine

class USBHandler:
    """
    Handler class for using USB Scan codes on a VirtualMachine (replaces VNC key codes).

    :param vm:
        A `VirtualMachine` object
    """
    def __init__(self, vm: 'VirtualMachine'):
        self._vm = vm
        self._client = vm._client
        # this prefix appears in front of all the keys
        self._keycode_name_prefix = 'KEY_'
        # contains a mapping of 'key as char/string' -> 'key as usb scan code (hex)'
        self._keycode_map = USBHandler._create_keycode_maps(keycode_name_prefix=self._keycode_name_prefix)
        # when true, the values in the keycodes.json file are in all caps (else assumes all lowercase)
        self._keycode_caps = True

    def _prep_key(self, key: str):
        """
        Handles:
        - if the key needs to be paired with the 'shift' key
        - finds the USB scan code for the provided string key
        - throws a UsbScanCodeError exception if the scan code can't be found
        - converts the found HEX code into an ESXi keyboard code and then into an ESXi key object
        - applies the 'shift' modifier if needed to the key object

        :param key:
            The key (as a string) to send through the ESXi API (e.g. 'a').
        :return:
            the key object to press
        """
        # Do we need to pair this key press with a shift modifier ?
        modifiers = None
        needs_shift = True if key.isupper() else False
        if key in '~!@#$%^&*()_+{}|:"<>?':
            needs_shift = True
            key = USBHandler._get_base_key(key)
        if needs_shift:
            # leftShift is preferable over right (short circuits conditional)
            modifiers = USBHandler._create_key_modifier(['leftShift'])
        
        # make the key's case match for map search
        key_char = key.upper() if self._keycode_caps else key.lower()

        if key_char == ' ':
            key_char = 'SPACE'

        # find this key in the map
        scan_code = self.get_usb_scan_code(key_char)
        
        # wrapper function that converts/shifts the keycode the way ESXi expects it and then
        # returns an API object representing the key
        return USBHandler._convert_usb_code_to_esxi_obj(scan_code, modifiers=modifiers)

    def press_key(self, key: str):
        """
        Uses the ESXi API to simulate a key press on the remote VM.
        For non-alphanumeric keys, use 'press_usb_code' instead.

        The key is sent as if a user was writing over a UI. It is up to the user to ensure that the VM
        is in a state to properly receive the key press (i.e. the console is open if trying to write to
        a console).

        :param key:
            The key (as a string) to send through the ESXi API (e.g. 'a').
        """
        # Do the prep work to convert this string into a key object
        key_obj = self._prep_key(key)
        log.debug(f"{str(self)} Simulating key press: {key}")
        # create the UsbScanCode API object and inform this VM to send the provided code
        self._send_keys_to_esxi([key_obj])

    def send_usb_code(self, keycode: str, modifier_names=[]):
        """
        Uses the ESXi API to simulate a key press on the remote VM.
        This function expects a HEX value (e.g. '0x04' for 'a').
        Use 'press_key' if you would like the HEX values converted for you.

        The key is sent as if a user was writing over a UI. It is up to the user to ensure that the VM
        is in a state to properly receive the key press (i.e. the console is open if trying to write to
        a console).

        :param keycode:
            The key to send through the ESXi API.
        :param modifier_names: 
            A list of strings containing the names of the keys you wish to apply to another keypress.
            If an invalid option is provided: will throw an 'UsbScanCodeModifierError' Exception
            Valid options are: 

                - leftAlt
                - leftControl
                - leftGui
                - leftShift
                - rightAlt
                - rightControl
                - rightGui
                - rightShift
        """
        # apply modifiers (such as leftShift)
        modifiers = USBHandler._create_key_modifier(modifier_names)

        # wrapper function that converts/shifts the keycode the way ESXi expects it and then
        # returns an API object representing the key
        key_obj = USBHandler._convert_usb_code_to_esxi_obj(keycode, modifiers=modifiers)

        # create the UsbScanCode API object and inform this VM to send the provided code
        log.debug(f"{str(self)} Sending key code: '{keycode}' with modifier names: {modifier_names}")
        self._send_keys_to_esxi([key_obj])

    def write(self, text: str, enter: bool = False):
        """
        Uses the ESXi API to simulate writing text on the remote VM.

        The keys are sent as if a user was writing over a UI. It is up to the user to ensure that the VM
        is in a state to properly receive the key presses (i.e. the console is open if trying to write to
        a console). 

        :param text:
            The text to write through USB scan codes.
        :param enter:
            Boolean value whether to send the `ENTER` key after writing the text.
        """
        keys_to_press = []
        for key in text:
            key_obj = self._prep_key(key)
            keys_to_press.append(key_obj)
        
        if enter:
            enter_key_obj = self._prep_key('ENTER')
            keys_to_press.append(enter_key_obj)

        # create the UsbScanCode API object and inform this VM to send the provided codes
        log.debug(f"{str(self)} Simulating write: {text}")
        self._send_keys_to_esxi(keys_to_press)
    
    def get_usb_scan_code(self, key_name: str) -> str:
        """
        Finds the scan code for the provided key_name (if one is known).
        Throws a UsbScanCodeError exception if no key is found with that name.

        :param key_name:
            the name of the key to search for
        :return:
            the USB Scan code as a HEX string
        """
        try:
            return self._keycode_map[key_name]
        except KeyError:
            raise exceptions.UsbScanCodeError(self._vm, f'Provided key name: "{key_name}" does not have a USB scan code mapping.')

    def get_key_name_by_scan_code(self, hex_scan_code: str) -> str:
        """
        Finds the name for the provided scan code (if one is known).
        Throws a UsbScanCodeError exception if no name is found with that code.

        :param hex_scan_code:
            the name of the USB scan code to search for (as HEX)
        :return:
            the name of the key with that scan code
        """
        try:
            return usbkeycodes.USB_CODE_TO_KEY_NAME[hex_scan_code].replace(self._keycode_name_prefix, '')
        except KeyError:
            raise exceptions.UsbScanCodeError(self._vm, f'Provided USB HEX scan code: "{hex_scan_code}" either does not exist or does not have a name.')

    def _send_keys_to_esxi(self, key_objects: typing.List) -> int:
        """
        Aggregates a list of KeyEvent objects into a UsbScanCode object.
        Sends the object to esxi to press the requested keys.

        :param key_objects:
            a list of key objects representing keys to be pressed on the VM
        :return:
            the number of keys injected
        """
        usbsc = pyVmomi.vim.vm.UsbScanCodeSpec()
        usbsc.keyEvents = key_objects
        keys_sent = self._vm._vim_vm.PutUsbScanCodes(usbsc)
        log.debug(f'Sent: "{keys_sent}" keys to the VM.')

    def __str__(self):
        return f"<{type(self).__name__} for {self._vm.name}>"
    
    def __repr__(self):
        return str(self)

    @staticmethod
    def _convert_usb_code_to_esxi_key_code(usb_hex_code: str):
        """
        Takes a USB scan code (represented in hex) and converts it to the cooresponding number in ESXi.
        Sets all lower bits to '1'.

        :param usb_hex_code:
            the usb scancode to convert (e.g. '0x04' for 'a')
        :return:
            the keycode as understood by the ESXi API
        """
        return (int(usb_hex_code, 16) << 16) | 7

    @staticmethod
    def _create_keycode_maps(keycode_name_prefix: str = 'KEY_') -> typing.Tuple[typing.Dict]:
        """
        Expects a JSON file containing a mapping of usb scan codes and their string values.
        Outputs two Python maps for use.

        :param keycode_prefix:
            any prefix to remove from the front of each key name in the dict
        :return:
            Tuple containing hex_to_str and str_to_hex dictionaries for keycodes from the provided filepath
        """
        str_to_hex = {str(v).replace(keycode_name_prefix, ''): k for k, v in usbkeycodes.USB_CODE_TO_KEY_NAME.items()}
        return str_to_hex
    
    @staticmethod
    def _get_key_event_obj(esxi_code: int, modifiers=None):
        """
        Converts an ESXi code value into the API object that represents a key.

        :param esxi_code:
            the esxi code integer to convert to an object
        :param modifiers:
            A UsbScanCodeSpecModifierType to attach to the key object (see _create_key_modifier)
        :return:
            the KeyEvent object used by the UsbScanCodeSpec object
        """
        ke = pyVmomi.vim.vm.UsbScanCodeSpec.KeyEvent()
        ke.usbHidCode = esxi_code
        if modifiers:
            ke.modifiers = modifiers
        return ke
    
    @staticmethod
    def _create_key_modifier(modifier_names: typing.List[str]):
        """
        Inputs a list of key names that you wish to apply as modifiers to a key and outputs an API object representing the modifiers.
        :param modifier_names: 
            A list of strings containing the names of the keys you wish to apply to another keypress.
            If an invalid option is provided: will throw an 'UsbScanCodeModifierError' Exception
            Valid options are: 

                - leftAlt
                - leftControl
                - leftGui
                - leftShift
                - rightAlt
                - rightControl
                - rightGui
                - rightShift
            
        :return:
            An API object representing all the modifiers applied
        """
        km = pyVmomi.vim.UsbScanCodeSpecModifierType()
        for m in modifier_names:
            m = m.strip().lower().replace(' ', '')
            if m == 'leftshift':
                km.leftShift = True
            elif m == 'rightshift':
                km.rightShift = True
            elif m == 'leftalt':
                km.leftAlt = True
            elif m == 'rightalt':
                km.rightAlt = True
            elif m == 'leftcontrol':
                km.leftControl = True
            elif m == 'rightcontrol':
                km.rightControl = True
            elif m == 'leftgui':
                km.leftGui = True
            elif m == 'rightgui':
                km.rightGui = True
            else:
                raise exceptions.UsbScanCodeModifierError(f'{m} is not a valid modifier type!')
        return km
    
    @staticmethod
    def _convert_usb_code_to_esxi_obj(usb_hex_code: str, modifiers=None):
        """
        Takes a USB scan code (represented in hex) and converts it into the API object that represents a key.

        :param usb_hex_code:
            the usb scancode to convert (e.g. '0x04' for 'a')
        :param modifiers:
            A UsbScanCodeSpecModifierType to attach to the key object (see _create_key_modifier)
        :return:
            the KeyEvent object used by the UsbScanCodeSpec object
        """
        return USBHandler._get_key_event_obj(USBHandler._convert_usb_code_to_esxi_key_code(usb_hex_code), modifiers=modifiers)
    
    @staticmethod
    def _get_base_key(key: str):
        """
        Gets the 'non-shift' character associated with a shifted character.

        :param key:
            the desired character
        :return:
            the base key/character on the keyboard
        """
        if key == '~':
            key = '`'
        elif key == '!':
            key = '1'
        elif key == '@':
            key = '2'
        elif key == '#':
            key = '3'
        elif key == '$':
            key = '4'
        elif key == '%':
            key = '5'
        elif key == '^':
            key = '6'
        elif key == '&':
            key = '7'
        elif key == '*':
            key = '8'
        elif key == '(':
            key = '9'
        elif key == ')':
            key = '0'
        elif key == '_':
            key = '-'
        elif key == '+':
            key = '='
        elif key == '{':
            key = '['
        elif key == '}':
            key = ']'
        elif key == '|':
            key = '\\'
        elif key == ':':
            key = ';'
        elif key == '"':
            key = '\''
        elif key == '<':
            key = ','
        elif key == '>':
            key = '.'
        elif key == '?':
            key = '/'
        return key



from esxi_utils.vm.virtualmachine import VirtualMachineList
from esxi_utils.util.connect.esxi import ESXiSSHConnection
from esxi_utils.networking.physicalnic import PhysicalNICList
from esxi_utils.networking.vmkernelnic import VMKernelNICList
from esxi_utils.networking.portgroup import PortGroupList
from esxi_utils.networking.distributedportgroup import DistributedPortGroupList
from esxi_utils.networking.vswitch import VSwitchList
from esxi_utils.networking.distributedvswitch import DistributedVSwitchList
from esxi_utils.datastore import DatastoreList
from esxi_utils.util import log, exceptions
from esxi_utils.firewall.firewall import Firewall
from pyVim.connect import SmartConnect, Disconnect
from contextlib import contextmanager
import threading
import pyVmomi
import typing
import time
import ssl

class ESXiClient:
	"""
	Client object for an ESXi host. Provides functionality for working with an ESXi host, such as querying
	for datastores, virtual machines, vswitches, and more.

	Typical usage of this package begins with initializing this client object and retrieving the desired objects
	through the provided API of this class.

	:param hostname: The hostname or IP of the ESXi server.
	:param username: The username for connecting over SSH.
	:param password: The password for connecting over SSH.
	:param child_hostname: When connecting to a vCenter instance, the hostname or IP of the child ESXi server.
	:param child_username: When connecting to a vCenter instance, the username to login to the child ESXi server (for SSH).
	:param child_password: When connecting to a vCenter instance, the password of the child ESXi user.
	"""
	def __init__(self, hostname: str, username: str, password: str, child_hostname: typing.Optional[str] = None, child_username: typing.Optional[str] = None, child_password: typing.Optional[str] = None):
		context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
		context.verify_mode = ssl.CERT_NONE
		err = None
		try:
			self._service_instance = SmartConnect(
				host=hostname,
				user=username, 
				pwd=password, 
				sslContext=context
			)
		except TimeoutError:
			err = TimeoutError(f"Connection timed out while attempting to connect to ESXi server at {hostname}. Check the correctness of the hostname and that the ESXi server is reachable.")
		except Exception as e:
			if "vim.fault.InvalidLogin" in str(e):
				err = ConnectionError(f"Cannot login to ESXi server at {hostname} as user \"{username}\" due to an incorrect user name or password.")
			else:
				err = e
		if err:
			raise err
		self._vcenter_hostname = hostname
		self._child_hostname = child_hostname
		self._child_username = child_username
		self._child_password = child_password
		self._host_system = None
		self._child_esxi_client_instance = None
		self._all_host_systems = self._get_vim_objects(pyVmomi.vim.HostSystem)
		self._host_system = self._find_host_system()
		if self._child_hostname and self._child_username and self._child_password:
			self._hostname = self._host_system.name
			try:
				self._child_esxi_client_instance = ESXiClient(self._child_hostname, self._child_username, self._child_password)
			except Exception as e:
				raise exceptions.ChildLoginFailureError(self._child_hostname, self._child_username)
		else:
			self._hostname = hostname
		self._username = username
		self._password = password
		self._keepalive = _ESXiClientKeepAlive(self)
		self._keepalive.start()
		log.info(f"{str(self)} Created.")
	
	def __enter__(self):
		return self
	
	def __exit__(self, *args):
		self.close()

	def close(self):
		"""
		Close the connection to this host.
		"""
		if hasattr(self, "_keepalive") and self._keepalive:
			self._keepalive.stop()
			self._keepalive.join()
		Disconnect(self._service_instance)
		if self._child_esxi_client_instance:
			self._child_esxi_client_instance.close()
	
	@property
	def hostname(self) -> str:
		"""
		The hostname for this ESXiClient.
		"""
		return self._hostname

	@property
	def username(self) -> str:
		"""
		The username for this ESXiClient.
		"""
		return self._username

	@property
	def password(self) -> str:
		"""
		The password for this ESXiClient.
		"""
		return self._password

	@property
	def vms(self) -> 'VirtualMachineList':
		"""
		The Virtual Machine handler for this host.
		"""
		return VirtualMachineList(self)

	@property
	def firewall(self) -> 'Firewall':
		"""
		The Firewall for this host.
		"""
		return Firewall(self)

	@property
	def datastores(self) -> 'DatastoreList':
		"""
		List of datastores on this host.
		"""
		return DatastoreList(self)
	
	@property
	def vswitches(self) -> 'VSwitchList':
		"""
		List of virtual switches on this host.
		"""
		return VSwitchList(self)
	
	@property
	def distributed_vswitches(self) -> 'DistributedVSwitchList':
		"""
		List of distributed virtual switches on this host.
		"""
		return DistributedVSwitchList(self)

	@property
	def portgroups(self) -> 'PortGroupList':
		"""
		List of port groups on this host.
		"""
		return PortGroupList(self)

	@property
	def distributed_portgroups(self) -> 'DistributedPortGroupList':
		"""
		List of distributed port groups on this host.
		"""
		return DistributedPortGroupList(self)
	
	@property
	def physicalnics(self) -> 'PhysicalNICList':
		"""
		List of physical NICs on this host.
		"""
		return PhysicalNICList(self)

	@property
	def vmkernelnics(self) -> 'VMKernelNICList':
		"""
		List of VMKernel NICs on this host.
		"""
		return VMKernelNICList(self)
	
	@property
	def hostNumCpuCores(self) -> int:
		"""
		The number of CPU cores that this ESXi server host has
		"""
		return self._host_system.hardware.cpuInfo.numCpuCores
	
	@property
	def hostCpuHz(self) -> int:
		"""
		The (clock?) speed of this ESXi server host's CPU cores
		"""
		return self._host_system.hardware.cpuInfo.hz
	
	def total_available_cpu_usage(self, unit: str = "MHz") -> float:
		"""
		Get the total available CPU capacity in the provided unit.
		The formula used for this is: (number of CPU cores total) * (core Speed in HZ) // ( 10^[3 *(unit_power)] )

		:param unit: The unit of measurement to use (one of: Hz, KHz, MHz, GHz).

		:return: The total amount of CPU capacity in the provided unit.
		"""
		hz_unit_orders = ["Hz", "KHz", "MHz", "GHz"]
		return self.hostNumCpuCores * self.hostCpuHz / 10**(3*hz_unit_orders.index(unit))

	def current_cpu_usage(self, unit: str = "MHz") -> float:
		"""
		Get the current CPU usage in the provided unit.

		:param unit: The unit of measurement to use (one of: Hz, KHz, MHz, GHz).

		:return: The current CPU usage in the provided unit.
		"""
		hz_unit_orders = ["Hz", "KHz", "MHz", "GHz"]
		return self._host_system.summary.quickStats.overallCpuUsage * 1000 * 1000 / 10**(3*hz_unit_orders.index(unit))
	
	def total_available_memory(self, unit: str = "B") -> float:
		"""
		Get the total amount of memory (RAM) installed on this ESXi host in the provided unit.
		
		:param unit: The unit of measurement to use (one of: B, KB, MB, GB, TB).

		:return: The total amount of memory (RAM) installed on this ESXi host.
		"""
		unit_orders = ["B", "KB", "MB", "GB", "TB"]
		assert unit in unit_orders, f"unit must be one of: {', '.join(unit_orders)}"
		# The API returns this in bytes
		return self._host_system.hardware.memorySize / pow(1024, unit_orders.index(unit))
	
	def current_memory_usage(self, unit: str = "B") -> float:
		"""
		Get the current amount of memory (RAM) in use on this ESXi host in the provided unit.
		
		:param unit: The unit of measurement to use (one of: B, KB, MB, GB, TB).

		:return: The current amount of memory (RAM) in use on this ESXi host.
		"""
		unit_orders = ["B", "KB", "MB", "GB", "TB"]
		assert unit in unit_orders, f"unit must be one of: {', '.join(unit_orders)}"
		# The API returns this quickStat in MB
		return self._host_system.summary.quickStats.overallMemoryUsage * 1024 * 1024 / pow(1024, unit_orders.index(unit))

	def memory_usage_percent(self) -> float:
		"""
		Returns the percentage usage of the RAM of the server
		"""
		return self.current_memory_usage() / self.total_available_memory()

	def cpu_usage_percent(self) -> float:
		"""
		Returns the percentage usage of the CPUs of the server 
		"""
		return self.current_cpu_usage() / self.total_available_cpu_usage()

	@contextmanager
	def ssh(self, force_parent: bool = False) -> typing.Generator['ESXiSSHConnection', None, None]:
		"""
		Connect to the underlying ESXi host over SSH. Yields a `UnixSSHConnection` object.

		:param force_parent:
			When connected to a vCenter system and this value is set to True:
			Will SSH into the vCenter hostname instead of the (default) child HostSystem hostname
			Authenication will fail if you do not have an account on the vCenter server.
		"""
		if force_parent:
			ip = self._vcenter_hostname
			user = self.username
			pw = self.password
		else:
			ip = self.hostname
			user = self._child_username if self._child_username is not None else self.username
			pw = self._child_password if self._child_password is not None else self.password
		with ESXiSSHConnection(ip=ip, username=user, password=pw) as conn:
			yield conn

	def __str__(self):
		return f"<{type(self).__name__} for {self.username}@{self.hostname}>"

	def __repr__(self):
		return str(self)

	def __del__(self):
		self.close()

	def _get_vim_objects_from(self, root, vim_type):
		"""
		Search the ESXi server for the all instances of the specified vim object under the given root object.

		:param root: The root object to search.
		:param vim_type: The vim type of the objects to get (e.g. pyVmomi.vim.VirtualMachine)

		:return: A list of all vim objects.
		"""
		if not isinstance(vim_type, list):
			vim_type = [vim_type]
		content = self._service_instance.RetrieveContent()
		container = content.viewManager.CreateContainerView(root, vim_type, True)
		objs = [ ref for ref in container.view ]
		container.Destroy()
		return objs

	def _get_vim_objects(self, vim_type, query_root: bool = False):
		"""
		Search the ESXi server for the all instances of the specified vim object.

		:param vim_type: The vim type of the objects to get (e.g. pyVmomi.vim.VirtualMachine)
		:param query_root: When True: Search from the 'root' Folder of ESXi instead of from the current 'host_system'

		:return: A list of all vim objects.
		"""
		# request only content based on the currently connected 'host_system' server
		if not query_root and self._child_hostname and self._host_system:
			return self._get_vim_objects_from(self._host_system, vim_type)
		return self._get_vim_objects_from(self._service_instance.RetrieveContent().rootFolder, vim_type)

	def _get_vim_object(self, vim_type, name: str):
		"""
		Search the ESXi server for the specified vim object.

		:param vim_type: The vim type of the object to get (e.g. pyVmomi.vim.VirtualMachine)
		:param name: The name of the object to get.

		:return: The vim object. Raises a RuntimeException if the object was not found, or if multiple we found.
		"""
		objs = self._get_vim_objects(vim_type)
		found = [ ref for ref in objs if ref.name == name ]
		if len(found) == 0:
			raise exceptions.ESXiAPIObjectNotFoundError(vim_type, f"{name} not found")
		elif len(found) > 1:
			raise exceptions.ESXiAPIObjectNotFoundError(vim_type, f"Multiple instances of {name} found")
		return found[0]

	def _get_network_object_from_host_system(self, name: str):
		"""
		Search the ESXi 'HostSystem' for the specified network object.
		This function specifically searches the HostSystem and not the 'root ServiceManager'.
		( Also see _get_vim_object() )

		:param name: The name of the object to get.

		:return: The vim object. Raises a RuntimeException if the object was not found, or if multiple we found.
		"""
		objs = self._get_network_objects_from_host_system()
		found = [ ref for ref in objs if ref.name == name ]
		if len(found) == 0:
			raise exceptions.ESXiAPIObjectNotFoundError(pyVmomi.vim.Network, f"{name} not found")
		elif len(found) > 1:
			raise exceptions.ESXiAPIObjectNotFoundError(pyVmomi.vim.Network, f"Multiple instances of {name} found")
		return found[0]

	def _get_network_objects_from_host_system(self):
		"""
		Returns the network objects associated with the connected 'HostSystem' object.

		:return: 'pyVmomi.VmomiSupport.ManagedObject[]' instance of managed network objects
		"""
		return self._host_system.network

	def _get_datastore_objects_from_host_system(self):
		"""
		Returns the datastore objects associated with the connected 'HostSystem' object.

		:return: 'pyVmomi.VmomiSupport.ManagedObject[]' instance of managed datastore objects
		"""
		return self._host_system.datastore

	def _wait_for_task(self, task):
		"""
		Wait for a `vim.Task` to complete.

		:param task: The `vim.Task` object to wait for.

		:return: The task result. If the task exits with status 'error', a exception will be thrown.
		"""
		assert isinstance(task, pyVmomi.vim.Task)
		while task.info.state not in [pyVmomi.vim.TaskInfo.State.error, pyVmomi.vim.TaskInfo.State.success]:
			time.sleep(0.2)
		if task.info.state == pyVmomi.vim.TaskInfo.State.error:
			raise exceptions.ESXiAPIError(str(task.info))
		return task.info.result

	def _find_host_system(self):
		"""
		Finds the correct 'HostSystem' to use based the 'child_hostname' param provided to init.
		If no 'child_hostname' is provided and there are multiple options, an ESXiAPIObjectNotFoundError is thrown.
		Otherwise, returns the only 'HostSystem' that exists.

		:return: the underlying host system to use
		"""
		# not a vCenter or no child hostname was specified
		if self._child_hostname is None:
			if isinstance(self._all_host_systems, list):
				if len(self._all_host_systems) > 1:
					raise exceptions.MultipleHostSystemsFoundError(self._all_host_systems)
				return self._all_host_systems[0]
			return self._all_host_systems
		# find the specified host system
		for esxi_server in self._all_host_systems:
			if self._child_hostname == esxi_server.name:
				return esxi_server
		# that hostname could not be found
		raise exceptions.ESXiAPIObjectNotFoundError('pyVmomi.vim.HostSystem', f"{self._child_hostname} child hostname not found")
# end ESXi client class

		
class _ESXiClientKeepAlive(threading.Thread):
	"""
	Class for keeping an ESXi client alive.

	:param client: An `ESXiClient` object
	"""
	def __init__(self, client):
		threading.Thread.__init__(self, daemon=True)
		assert isinstance(client, ESXiClient)
		self.service_instance = client._service_instance
		self.id = str(client)
		self.running = True

	def stop(self):
		"""
		Stop the keep-alive thread.
		"""
		self.running = False
	
	def run(self):
		i = 0
		while self.running:
			time.sleep(1)
			i += 1
			if i == 300:
				self.service_instance.CurrentTime()
				i = 0
			


from esxi_utils.util import log, exceptions
from lxml import etree
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.firewall.firewall import Firewall
	from esxi_utils.firewall.rule import RuleList

class RulesetList:
	"""
	List of Rulesets for the host firewall.
	"""
	def __init__(self, firewall: 'Firewall'):
		self._firewall = firewall

	def __iter__(self) -> typing.Iterator['Ruleset']:
		return iter([ Ruleset(self._firewall, obj.key) for obj in self._firewall._firewall_info.ruleset ])

	def __getitem__(self, name) -> 'Ruleset':
		return self.get(name)

	def __contains__(self, name: str) -> bool:
		return self.exists(name)

	@property
	def items(self) -> typing.List['Ruleset']:
		"""
		A list of all items
		"""
		return list(self)

	def find(self, key: str) -> typing.Union['Ruleset', None]:
		"""
		Get a ruleset by key.

		:param key: The ruleset key to search for.

		:return: The `Ruleset` object, or `None` if not found.
		"""
		found = [ ruleset for ruleset in self if ruleset.key == key ]
		if len(found) == 0:
			return None
		return found[0]

	def get(self, key: str) -> 'Ruleset':
		"""
		Get a ruleset by key and raise an exception if not found.

		:param key: The ruleset key to search for.

		:return: The `Ruleset` object
		"""
		ruleset = self.find(key)
		if ruleset is None:
			raise exceptions.FirewallError(self, f"No ruleset found with key '{key}'")
		return ruleset

	def exists(self, key: str) -> bool:
		"""
		Check whether a ruleset exists.

		:parma key: The ruleset key

		:return: Whether or not the ruleset exists
		"""
		return self.find(key) is not None

	def add(self, key: str, enabled: str = True, required: str = False) -> 'Ruleset':
		"""
		Add a firewall ruleset. If this ruleset already exists, an error will be raised.

		:param key: The new ruleset's key.
		:param enabled: Whether or not this ruleset should be enabled.
		:param required: Whether or not this ruleset is required.

		:return: A `Ruleset` object for the new ruleset.
		"""
		assert isinstance(key, str), "key must be a string"
		assert isinstance(enabled, bool), "enabled must be a boolean"
		assert isinstance(required, bool), "required must be a boolean"

		log.info(f"{str(self)} Adding new ruleset: {key}")
		if self.exists(key):
			raise exceptions.FirewallError(self, f"A ruleset already exists with key '{key}'")

		tree = self._firewall._get_service_xml()

		# Find the next available number ID for the ruleset
		# Rulesets are numbered by a unique 4-digit ID (in addition to their "text" ID)
		# We find the highest currently on the firewall and then add 1
		ids = []
		for node in tree.findall("./service"):
			ids.append(int(node.get("id")))
		max_id = max(ids)
		if max_id >= 9999:
			raise exceptions.FirewallError(self, f"Maximum ruleset ID reached ({max_id})")
		next_id = str(max_id+1).zfill(4)
		log.debug(f"Next available ruleset ID: {next_id}")

		# Create the service node and add comments
		service_tag = etree.SubElement(tree, "service")
		service_tag.set("id", next_id)
	
		# Add the "text" id tag to this ruleset
		id_tag = etree.SubElement(service_tag, "id")
		id_tag.text = key
		
		# Add remaining tags to ruleset
		enabled_tag = etree.SubElement(service_tag, "enabled")
		enabled_tag.text = "true" if enabled else "false"

		required_tag = etree.SubElement(service_tag, "required")
		required_tag.text = "true" if required else "false"

		self._firewall._update_service_xml(tree)
		# on vCenter the connection may get messed up during firewall refresh.
		# try to find the key in the child server
		if self._firewall._client._child_hostname:
			log.error('WARN: Enabling VNC in a vCenter instance may cause communication errors! Manually reset the server to green and "connect" it in vCenter.')
			return self._firewall._client._child_esxi_client_instance.firewall.rulesets.get(key)
		return self.get(key)

	def enable_vnc(self) -> 'Ruleset':
		"""
		Add a ruleset for VNC access.

		:return: The created VNC ruleset.
		"""
		ruleset = self.add("VNC", enabled=True, required=False)
		ruleset.rules.add(direction="inbound", protocol="tcp", porttype="dst", begin=5900, end=6000)
		return ruleset

	def __str__(self):
		return f"<{type(self).__name__} for {self._firewall._client.hostname} ({len(self.items)} rulesets)>"

	def __repr__(self):
		return str(self)


class Ruleset:
	"""
	Class for an ESXi firewall ruleset.
	"""
	def __init__(self, firewall: 'Firewall', key: str):
		self._firewall = firewall
		self._key = key

	@property
	def key(self) -> str:
		"""
		The ruleset key.
		"""
		return self._key

	@property
	def label(self) -> str:
		"""
		Display label for the ruleset.
		"""
		return self._obj.label

	@property
	def service(self) -> str:
		"""
		Managed service (if any) that uses this ruleset.
		"""
		return self._obj.service

	@property
	def enabled(self) -> bool:
		"""
		Flag indicating whether the ruleset is enabled.
		"""
		return self._obj.enabled

	@property
	def rules(self) -> 'RuleList':
		"""
		List of rules within the ruleset.
		"""
		from esxi_utils.firewall.rule import RuleList
		return RuleList(self)

	def remove(self):
		"""
		Remove this ruleset from its parent firewall.
		"""
		tree, ruleset = self._get_xml_ruleset()
		ruleset.getparent().remove(ruleset)
		self._firewall._update_service_xml(tree)

	def _get_xml_ruleset(self):
		"""
		Get this ruleset from /etc/vmware/firewall/service.xml as an `etree.Element` object.

		:return: A tuple (tree, element)
		"""
		tree = self._firewall._get_service_xml()
		found = tree.find(f"./service[id='{self.key}']") 
		if found is None:
			raise exceptions.FirewallError(self, f"Failed to find ruleset in /etc/vmware/firewall/service.xml")
		return (tree, found)

	@property
	def _obj(self):
		for obj in self._firewall._firewall_info.ruleset:
			if obj.key == self.key:
				return obj
		raise exceptions.FirewallError(self, f"No ruleset found with key '{self.key}'")

	def __str__(self):
		return f"<Ruleset key='{self.key}' label='{self.label}' service='{self.service}' enabled='{self.enabled}' ({len(self.rules.items)} rules)>"

	def __repr__(self):
		return str(self)


from esxi_utils.util.connect.cisco import CiscoSSHConnection
from esxi_utils.util.connect.esxi import ESXiSSHConnection
from esxi_utils.util.connect.panos import PanosAPIConnection, PanosSSHConnection
from esxi_utils.util.connect.ssh import SSHConnection, SSHResponsePromise
from esxi_utils.util.connect.unix import UnixSSHConnection
from esxi_utils.util.connect.winrm import WinRMConnection

__all__ = [
	"CiscoSSHConnection",
	"ESXiSSHConnection",
	"PanosAPIConnection",
	"PanosSSHConnection",
	"SSHConnection",
    "SSHResponsePromise",
	"UnixSSHConnection",
	"WinRMConnection"
]


from esxi_utils.vm.virtualmachine import VirtualMachine
from esxi_utils.vm.tools.panos import PanosGuestTools
from esxi_utils.util.connect.panos import PanosAPIConnection, PanosSSHConnection
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.vm.types.ostype import OSType

class PaloAltoFirewallVirtualMachine(VirtualMachine):
	"""
	A Palo Alto Firewall virtual machine on an ESXi host. 
	"""
	@property
	def api(self) -> typing.Type['PanosAPIConnection']:
		"""
		A connection object for establishing a remote connection to this Palo Alto Firewall's API.
		"""
		return PanosAPIConnection

	@property
	def ssh(self) -> typing.Type['PanosSSHConnection']:
		"""
		A SSH connection object for establishing a remote connection to this Palo Alto Firewall.
		"""
		return PanosSSHConnection

	@property
	def tools(self) -> 'PanosGuestTools':
		"""
		Get the VMware Tools object for this VM. Includes additional methods for interaction with the Palo Alto CLI.
		"""
		return PanosGuestTools(self)
	
	@property
	def ostype(self) -> 'OSType':
		"""
		The `OSType` for this Virtual Machine.
		"""
		from esxi_utils.vm.types.ostype import OSType
		return OSType.PanOs
	


from esxi_utils.util import log, exceptions
import pyVmomi
import os
import re
import urllib.request
import urllib.parse
import typing
import ssl

if typing.TYPE_CHECKING:
	from esxi_utils.client import ESXiClient
	from esxi_utils.vm.virtualmachine import VirtualMachine

class DatastoreList:
	"""
	The collection of datastores on an ESXi host.

	:param client: The `ESXiClient` object for the ESXi server to use.
	"""
	def __init__(self, client: 'ESXiClient'):
		self._client = client

	def __iter__(self) -> typing.Iterator['Datastore']:
		datastores = []
		# The host system keeps a record of datastore objects, but I didn't find a reference to the datastore
		# Ideally, we could get the datastore ID from the '_host_system' object instead of having to iterate
		if self._client._child_hostname:
			client_moIds = []
			for ds in self._client._get_datastore_objects_from_host_system():
				client_moIds.append(ds._moId)
		# iterate through all datastores
		for dc in self._client._get_vim_objects(pyVmomi.vim.Datacenter, query_root=True):
			for ds in dc.datastore:
				# filter out datastores not relevant to this '_host_system'
				if self._client._child_hostname and ds._moId not in client_moIds:
					continue
				datastores.append(Datastore(self._client, dc._moId, ds._moId))
		return iter(datastores)

	def __getitem__(self, name) -> 'Datastore':
		return self.get(name)

	def __contains__(self, name: str):
		return self.exists(name)

	@property
	def items(self) -> typing.List['Datastore']:
		"""
		A list of all items
		"""
		return list(self)

	def find(self, name: str) -> typing.Union['Datastore', None]:
		"""
		Get a datastore on the ESXi host.

		:param name: The name of the datastore to get.

		:return: A `Datastore` object, or `None` if not found.
		"""
		ds = [ds for ds in self if ds.name == name]
		if len(ds) == 0:
			return None
		return ds[0]

	def get(self, name: str) -> 'Datastore':
		"""
		Get a datastore on the ESXi host and raise an exception if not found.

		:param name: The name of the datastore to get.

		:return: A `Datastore` object
		"""
		ds = self.find(name)
		if ds is None:
			raise exceptions.DatastoreNotFoundError(name)
		return ds

	@property
	def names(self) -> typing.List[str]:
		"""
		List the names of all datastores on the ESXi host.

		:return: A list of names (strings).
		"""
		return [ ds.name for ds in self ]

	def exists(self, name: str) -> bool:
		"""
		Return whether or not the given datastore exists.

		:param name: The name of the datastore.

		:return: A boolean whether the datastore exists.
		"""
		return name in self.names

	def resolve(self, name_or_obj: typing.Union[str, 'Datastore']) -> 'Datastore':
		"""
		Resolve a type to an `Datastore` object. Raises an error if the datastore cannot be resolved.

		:param name_or_obj: The value to resolve. If a string, an `Datastore` is searched for by name. If already an `Datastore` object, this simply asserts the datastore exists.

		:return: An `Datastore` object.
		"""
		if isinstance(name_or_obj, str):
			name_or_obj = self.get(name=name_or_obj)
		if not isinstance(name_or_obj, Datastore) or not self.exists(name_or_obj.name):
			raise exceptions.DatastoreNotFoundError(name_or_obj)
		return name_or_obj

	def __str__(self):
		return f"<{type(self).__name__} for {self._client.username}@{self._client.hostname}>"

	def __repr__(self):
		return str(self)


class Datastore:
	"""
	A datastore on the ESXi host.
	"""
	def __init__(self, client: 'ESXiClient', datacenter_id: str, datastore_id: str):
		self._client = client
		self._datacenter_id = datacenter_id
		self._datastore_id = datastore_id

	@property
	def name(self) -> str:
		"""
		Get the name of the this datastore.

		:return: The datastore name.
		"""
		return self._datastore.summary.name

	@property
	def file_system_type(self) -> str:
		"""
		Get the type of file system volume, such as VMFS or NFS.

		:return: The file system type
		"""
		return self._datastore.summary.type

	@property
	def nfs(self) -> bool:
		"""
		Return whether or not this datastore is a network file system.

		:return: A boolean whether the file system type is NFS.
		"""
		return self.file_system_type == "NFS"

	@property
	def path(self) -> str:
		"""
		Get the path of the datastore on the ESXi host (i.e. /vmfs/volumes/<id>).

		:return: The path of the datastore
		"""
		return self._datastore.summary.url

	@property
	def accessible(self) -> bool:
		"""
		Get the connectivity status of this datastore. If false, this means the datastore is not accessible and certain properties (i.e. this datastore's capacity and freespace properties) cannot be validated.

		:return: Whether or not the datastore is accessible.
		"""
		return self._datastore.summary.accessible

	@property
	def root(self) -> 'DatastoreFile':
		"""
		Get the root filepath for this datastore.
		"""
		return DatastoreFile(self, ".")

	@property
	def vms(self) -> typing.List['VirtualMachine']:
		"""
		Virtual machines stored on this datastore. 
		"""
		vm_ids = [ str(vm._moId) for vm in self._datastore.vm ]
		return [ vm for vm in self._client.vms.items if vm.id in vm_ids]

	def filepath(self, path: str) -> 'DatastoreFile':
		"""
		Get a `DatastoreFile` object for the provided path.
		"""
		return DatastoreFile(self, path)

	def capacity(self, unit: str = "B") -> int:
		"""
		Get the capacity of this datastore in the provided unit.
		
		:param unit: The unit of measurement to use (one of: B, KB, MB, GB, TB).

		:return: The capacity of the datastore in the provided unit.
		"""
		unit_orders = ["B", "KB", "MB", "GB", "TB"]
		assert unit in unit_orders, f"unit must be one of: {', '.join(unit_orders)}"
		return self._datastore.summary.capacity / pow(1024, unit_orders.index(unit))

	def freespace(self, unit: str = "B") -> int:
		"""
		Get the free space of this datastore in the provided unit.
		
		:param unit: The unit of measurement to use (one of: B, KB, MB, GB, TB).

		:return: The free space of the datastore in the provided unit.
		"""
		unit_orders = ["B", "KB", "MB", "GB", "TB"]
		assert unit in unit_orders, f"unit must be one of: {', '.join(unit_orders)}"
		return self._datastore.summary.freeSpace / pow(1024, unit_orders.index(unit))

	def used_disk_space(self, unit: str = "B") -> int:
		"""
		Get the used space of this datastore in the provided unit.
		
		:param unit: The unit of measurement to use (one of: B, KB, MB, GB, TB).

		:return: The used space of the datastore in the provided unit.
		"""
		unit_orders = ["B", "KB", "MB", "GB", "TB"]
		assert unit in unit_orders, f"unit must be one of: {', '.join(unit_orders)}"
		return self.capacity(unit) - self.freespace(unit)

	def disk_usage_percent(self) -> float:
		"""
		Returns the percentage usage of the disk space of this drive
		"""
		return self.used_disk_space() / self.capacity()

	@property
	def _datacenter(self):
		dc = pyVmomi.vim.Datacenter(self._datacenter_id)
		dc._stub = self._client._service_instance._stub
		return dc

	@property
	def _datastore(self):
		ds = pyVmomi.vim.Datastore(self._datastore_id)
		ds._stub = self._client._service_instance._stub
		return ds

	def __str__(self):
		return f"<{type(self).__name__} '{self.name}' on {self._client.hostname}>"

	def __repr__(self):
		return str(self)


class DatastoreFile:
	"""
	Class representing a filepath on a datastore. 
	It is possible to navigate using the div operator (i.e. `filepath = filepath / "path" / "to" / "file"`) or by indexing (i.e. `filepath = filepath["path"]["to"]["file"]`)

	:param datastore: The `Datastore` object.
	:param relpath: The relative path to the file on the datastore.
	"""
	def __init__(self, datastore: 'Datastore', relpath: str):
		self._datastore = datastore
		self._relpath = self._sanitize(relpath)

	def	__truediv__(self, path: str) -> 'DatastoreFile':
		return self.join(path)

	def __getitem__(self, path) -> 'DatastoreFile':
		return self.join(path)

	def __iter__(self):
		return iter(self.ls(recursive=False))

	def __contains__(self, name: str):
		return self.join(name).exists

	def join(self, *paths) -> 'DatastoreFile':
		"""
		Join this path with one or more additional path components. Does not check that this path exists.
		"""
		return DatastoreFile(self.datastore, os.path.join(self.relpath, *paths))

	@property
	def datastore(self) -> 'Datastore':
		"""
		The datastore for this filepath.
		"""
		return self._datastore

	@property
	def relpath(self) -> str:
		"""
		Return the relative path for this filepath in the datastore.
		"""
		return self._relpath

	@property
	def path(self) -> str:
		"""
		Return the path for this filepath in the datastore.
		"""
		return f"[{self.datastore.name}] {self.relpath}"

	@property
	def abspath(self) -> str:
		"""
		The absolute path for this filepath in the datastore.
		"""
		return os.path.join(self.datastore.path, self.relpath)

	@property
	def filename(self) -> str:
		"""
		The final component of the file path (basename).
		"""
		return os.path.basename(self.relpath)

	@property
	def parent(self) -> typing.Union['DatastoreFile', None]:
		"""
		The parent directory of this file. If there is no parent, this returns `None`.
		"""
		if self.relpath == ".":
			return None
		return DatastoreFile(self.datastore, os.path.dirname(self.relpath))

	def _get_files(self, recursive=False):
		"""
		List files in the datastore.

		:param recursive: Whether or not to list files recursively (i.e. list path and all subfolders).

		:return: A dict mapping file paths to metadata
		"""
		log.debug(f"<{self.path}> Getting files (recursive={recursive})...")
		spec = pyVmomi.vim.host.DatastoreBrowser.SearchSpec(details=pyVmomi.vim.host.DatastoreBrowser.FileInfo.Details(fileSize=True))
		task = None
		if recursive:
			task = self.datastore._datastore.browser.SearchDatastoreSubFolders_Task(datastorePath=self.path, searchSpec=spec)	
		else:
			task = self.datastore._datastore.browser.SearchDatastore_Task(datastorePath=self.path, searchSpec=spec)

		results = self.datastore._client._wait_for_task(task)
		if not isinstance(results, list):
			results = [results]
		paths = {}
		base = f"[{self.datastore.name}]"
		for result in results:
			folder_path = result.folderPath.replace(base, "", 1).strip()
			for file in result.file:
				path = os.path.join(folder_path, file.path)
				path = self._sanitize(path)
				paths[path] = {
					"size": file.fileSize,
					"isfile": not isinstance(file, pyVmomi.vim.host.DatastoreBrowser.FolderInfo),
				}
		return paths

	def ls(self, recursive: bool = False) -> typing.List['DatastoreFile']:
		"""
		List all files and directories in this directory.

		:param: List files recursively.

		:return: A list of `DatastoreFile` objects
		"""
		return [ DatastoreFile(self.datastore, path) for path in self._get_files(recursive=recursive).keys() ]

	@property
	def files(self) -> typing.List['DatastoreFile']:
		"""
		List files in this directory.
		"""
		return [ DatastoreFile(self.datastore, path) for path, stat in self._get_files(recursive=False).items() if stat["isfile"] ]

	@property
	def dirs(self) -> typing.List['DatastoreFile']:
		"""
		List directories in this directory.
		"""
		return [ DatastoreFile(self.datastore, path) for path, stat in self._get_files(recursive=False).items() if not stat["isfile"] ]

	@property
	def stat(self) -> typing.Dict[str, typing.Any]:
		"""
		Information on this file or directory. Raises a `DatastoreFileNotFoundError` exception if this file does not exist
		"""
		log.debug(f"<{self.path}> Stat...")
		if self.parent is None:
			result = { "size": 0, "isfile": False }
			log.debug(f"<{self.path}> Stat result (root): {result}")
			return result
		try:
			folder_contents = self.parent._get_files(recursive=False)
		except exceptions.ESXiAPIError as e:
			if "vim.fault.FileNotFound" in str(e):
				raise exceptions.DatastoreFileNotFoundError(self.datastore, self.relpath)
			log.warn(f'Exception getting folder_contents in stat: {str(e)}')				
		if self.relpath not in folder_contents:
			raise exceptions.DatastoreFileNotFoundError(self.datastore, self.relpath)
		result = folder_contents[self.relpath]
		log.debug(f"<{self.path}> Stat result: {result}")
		return result

	@property
	def exists(self) -> bool:
		"""
		Whether or not this file exists on the datastore.
		"""
		try:
			return self.stat is not None
		except exceptions.DatastoreFileNotFoundError:
			return False

	@property
	def isfile(self) -> bool:
		"""
		Whether or not this is a file on the datastore.
		"""
		try:
			return self.stat["isfile"]
		except exceptions.DatastoreFileNotFoundError:
			return False

	@property
	def isdir(self) -> bool:
		"""
		Whether or not this is a directory on the datastore.
		"""
		try:
			return not self.stat["isfile"]
		except exceptions.DatastoreFileNotFoundError:
			return False

	def read(self, encoding: typing.Optional[str] = 'utf-8') -> typing.Union[str, bytes]:
		"""
		Returns the contents of a file on the remote server.

		:param encoding: The encoding to use when decoding the file. If `None`, no decoding is performed and bytes will be returned.

		:return: The file contents as a string (or bytes).
		"""
		log.debug(f"<{self.path}> Reading (encoding='{encoding}')...")
		if not self.isfile:
			raise exceptions.DatastoreIsADirectoryError(self.datastore, self.relpath)

		params = {"dsName": self.datastore._datastore.info.name, "id": self.datastore._datacenter._moId}
		query_string = urllib.parse.urlencode(params)
		url = "https://" + self.datastore._client.hostname + ":443/folder/" + urllib.parse.quote_plus(self.relpath) + "?" + query_string

		# Download the file
		ctx = ssl.create_default_context()
		ctx.check_hostname = False
		ctx.verify_mode = ssl.CERT_NONE
		try:
			request = urllib.request.Request(url)
			request.add_header("Content-Type", "application/octet-stream")
			child_client = self.datastore._client._child_esxi_client_instance
			cookie = self.datastore._client._service_instance._stub.cookie if child_client is None else child_client._service_instance._stub.cookie
			request.add_header("Cookie", cookie)

			with urllib.request.urlopen(request, context=ctx) as response:
				response_content = response.read()
				if not response.status == 200:
					raise exceptions.DatastoreError(self.datastore, f"Error while getting file {self.relpath} (status {response.status}): {response_content}")
				if encoding:
					return response_content.decode(encoding=encoding)
				return response_content
		except (urllib.error.URLError, urllib.error.HTTPError) as error:
			raise exceptions.DatastoreError(self.datastore, f"Error while getting file {self.relpath}: {error.reason}")

	def write(self, contents: typing.Union[str, bytes]):
		"""
		Writes to the remote file. The file will be created if it does not exist. If it does exist, the contents will be overwritten.

		:param path: The path to a file, relative to this datastore.
		:param contents: The contents to write to the file. 
		"""
		log.debug(f"<{self.path}> Writing (len={len(contents)})...")
		if not self.isfile:
			raise exceptions.DatastoreIsADirectoryError(self.datastore, self.relpath)

		params = {"dsName": self.datastore._datastore.info.name, "id": self.datastore._datacenter._moId}
		query_string = urllib.parse.urlencode(params) 
		url = "https://" + self.datastore._client.hostname + ":443/folder/" + urllib.parse.quote_plus(self.relpath) + "?" + query_string

		if isinstance(contents, str):
			contents = str.encode(contents)

		# Upload the file
		ctx = ssl.create_default_context()
		ctx.check_hostname = False
		ctx.verify_mode = ssl.CERT_NONE
		try:
			request = urllib.request.Request(url, data=contents, method='PUT')
			request.add_header("Content-Type", "application/octet-stream")
			child_client = self.datastore._client._child_esxi_client_instance
			cookie = self.datastore._client._service_instance._stub.cookie if child_client is None else child_client._service_instance._stub.cookie
			request.add_header("Cookie", cookie)

			with urllib.request.urlopen(request, context=ctx) as response:
				response_content = response.read()
				if not response.status >= 200 and not response.status < 300:
					raise exceptions.DatastoreError(self.datastore, f"Error while writing file {self.relpath} (status {response.status}): {response_content}")
		except (urllib.error.URLError, urllib.error.HTTPError) as error:
			raise exceptions.DatastoreError(self.datastore, f"Error while writing file {self.relpath}: {error.reason}")

	def download(self, dst: str, directory_contents_only: bool = False, overwrite: bool = False) -> typing.List[str]:
		"""
		Download this file or directory from the datastore.

		:param dst: The local destination file or directory to download to.
		:param directory_contents_only: If `True` and this is a directory, then just the contents of the directory will be downloaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files.

		:return: A list of local paths to all files downloaded.
		"""
		log.debug(f"<{self.path}> Downloading to \"{dst}\"...")
		dst = os.path.abspath(dst)
		if not self.exists:
			raise exceptions.DatastoreFileNotFoundError(self.datastore, self.relpath)

		download_map = {} # Map download source -> destination
		if self.isfile:
			# Single file download
			target = dst
			parent = os.path.dirname(target)
			if not os.path.isdir(parent):
				raise NotADirectoryError(f"{parent} is not a valid local directory")
			# target exists
			if os.path.exists(target):
				# complete the path to the file if the input is just the directory
				if os.path.isdir(target):
					target = os.path.join(target, self.filename)
				# else: assume the target is already the full path to the file
			download_map[self.relpath] = target
		else:
			# Directory download
			target = dst
			if os.path.isdir(target) and directory_contents_only:
				# Downloading directory contents into a directory
				pass
			elif os.path.isdir(target) and not directory_contents_only:
				# Downloading directory into a directory, download with the same name into the target
				target = os.path.join(target, self.filename)
			elif not os.path.exists(target) and directory_contents_only:
				# We cannot download the contents of a directory if the dst does not exist
				raise NotADirectoryError(f"{dst} is not a valid local directory")
			elif not os.path.exists(target) and not directory_contents_only:
				# Downloading directory under a new name, we just need to ensure that the parent directory exists
				parent = os.path.dirname(target)
				if not os.path.isdir(parent):
					raise NotADirectoryError(f"{parent} is not a valid local directory")
			
			download_map = { file: os.path.join(target, file[len(self.relpath):].strip("/")) for file, stat in self._get_files(recursive=True).items() if stat["isfile"] }

		if len(download_map.keys()) == 0:
			return []

		for target in download_map.values():
			if os.path.exists(target) and not overwrite:
				raise FileExistsError(f"{target} already exists")
			if os.path.isdir(target):
				raise IsADirectoryError(f"{target} cannot be overwritten as it exists as a directory")

		# Begin downloading files
		params = {"dsName": self.datastore._datastore.info.name, "id": self.datastore._datacenter._moId}
		query_string = urllib.parse.urlencode(params) 

		ctx = ssl.create_default_context()
		ctx.check_hostname = False
		ctx.verify_mode = ssl.CERT_NONE

		for remotepath, localpath in download_map.items():
			log.debug(f"<{self.path}> Downloading {remotepath} to {localpath}")
			url = "https://" + self.datastore._client.hostname + ":443/folder/" + urllib.parse.quote_plus(remotepath) + "?" + query_string
			request = urllib.request.Request(url)
			request.add_header("Content-Type", "application/octet-stream")
			child_client = self.datastore._client._child_esxi_client_instance
			cookie = self.datastore._client._service_instance._stub.cookie if child_client is None else child_client._service_instance._stub.cookie
			request.add_header("Cookie", cookie)
			os.makedirs(os.path.dirname(localpath), exist_ok=True)
			try:
				with urllib.request.urlopen(request, context=ctx) as response, open(localpath, 'wb') as out_file:
					if not response.status == 200:
						raise exceptions.DatastoreError(self.datastore, f"Error while downloading file {remotepath} (status {response.status}): {response.read()}")
					while True:
						content = response.read(256 * 1024)
						if len(content) == 0:
							# Done
							break
						out_file.write(content)
					out_file.flush()
					os.fsync(out_file.fileno())
			except (urllib.error.URLError, urllib.error.HTTPError) as error:
				raise exceptions.DatastoreError(self.datastore, f"Error while downloading file {remotepath}: {error.reason}")
		return list(download_map.values())

	def upload(self, src: str, directory_contents_only: bool = False, overwrite: bool = False) -> typing.List['DatastoreFile']:
		"""
		Upload a file or directory to this path on the datastore.

		:param src: The local path to a file or directory to upload.
		:param directory_contents_only: If `True` and `src` points to a directory, then only the contents of the directory will be uploaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files on the datastore.

		:return: A list of `DatastoreFile` objects for all files uploaded.
		"""
		log.info(f"<{self.path}> Uploading \"{src}\"...")
		src = os.path.abspath(src)
		if not os.path.exists(src):
			raise FileNotFoundError(f"File or directory {src} not found")

		upload_map = {} # Map upload source -> destination
		if os.path.isfile(src):
			# Single file upload
			target = self
			if not target.exists:
				# Target doesn't exist, we will upload under the name specified
				# Just ensure the parent directory does exist
				if not target.parent.isdir:
					raise exceptions.DatastoreNotADirectoryError(self.datastore, target.parent.relpath)
			else:
				# Target exists, we'll upload into the target with the same name
				if not target.isdir:
					raise exceptions.DatastoreNotADirectoryError(self.datastore, target.relpath)
				target = target / os.path.basename(src)
			upload_map[src] = target
		elif os.path.isdir(src):
			# Directory upload
			target = self
			target_exists = target.exists
			target_is_dir = target.isdir
			if target_is_dir and directory_contents_only:
				# Uploading directory contents into a directory
				pass
			elif target_is_dir and not directory_contents_only:
				# Uploading directory into a directory, upload with the same name into the target
				target = target / os.path.basename(src)
			elif not target_exists and directory_contents_only:
				# We cannot upload the contents of a directory if the dst does not exist
				raise exceptions.DatastoreNotADirectoryError(self.datastore, target.relpath)
			elif not target_exists and not directory_contents_only:
				# Uploading directory under a new name, we just need to ensure that the parent directory exists
				if not target.parent.isdir:
					raise exceptions.DatastoreNotADirectoryError(self.datastore, target.parent.relpath)
			
			for root, _, files in os.walk(src):
				for file in files:
					path = os.path.join(root, file)
					upload_map[path] = target / path[len(src):].strip("/")

		if len(upload_map.keys()) == 0:
			return []

		for target in upload_map.values():
			if target.exists and not overwrite:
				raise exceptions.DatastoreFileExistsError(self.datastore, target.relpath)
			if target.isdir:
				raise exceptions.DatastoreIsADirectoryError(self.datastore, target.relpath)

		# Begin uploading files
		params = {"dsName": self.datastore._datastore.info.name, "id": self.datastore._datacenter._moId}
		query_string = urllib.parse.urlencode(params) 

		ctx = ssl.create_default_context()
		ctx.check_hostname = False
		ctx.verify_mode = ssl.CERT_NONE

		for localpath, remotepath in upload_map.items():
			log.debug(f"<{self.path}> Uploading {localpath} to {remotepath.relpath}")
			url = "https://" + self.datastore._client.hostname + ":443/folder/" + urllib.parse.quote_plus(remotepath.relpath) + "?" + query_string

			with open(localpath, mode='rb') as file:
				request = urllib.request.Request(url, data=file, method='PUT')
				request.add_header("Content-Type", "application/octet-stream")
				child_client = self.datastore._client._child_esxi_client_instance
				cookie = self.datastore._client._service_instance._stub.cookie if child_client is None else child_client._service_instance._stub.cookie
				request.add_header("Cookie", cookie)
				remotepath.parent.mkdir(parents=True)
				try:
					with urllib.request.urlopen(request, context=ctx) as response: # if you add timeout=#, it does timeout
						response_content = str(response.read())
						if not response.status >= 200 and not response.status < 300:
							raise exceptions.DatastoreError(self.datastore, f"Error while uploading file {localpath} (status {response.status}): {response_content}")
				except (urllib.error.URLError, urllib.error.HTTPError) as error:
					raise exceptions.DatastoreError(self.datastore, f"Error while uploading file {localpath}: {error.reason}")

		return list(upload_map.values())

	def mkdir(self, parents: bool = False):
		"""
		Create this folder (directory) in the datastore. If the directory already exists, this does nothing.

		:param parents: Make parent directories as needed.
		"""
		log.debug(f"<{self.path}> Mkdir...")
		try:
			file_manager = self.datastore._client._service_instance.RetrieveContent().fileManager
			file_manager.MakeDirectory(name=self.path, datacenter=self.datastore._datacenter, createParentDirectories=parents)
		except pyVmomi.vim.fault.FileAlreadyExists:
			pass

	def copy(self, to: 'DatastoreFile', force: bool = False):
		"""
		Copy this file or directory to another location.

		:param to: The `DatastoreFile` to the destination file or directory.
		:param force: If true, overwrite any identically named file at the destination.
		"""
		assert isinstance(to, DatastoreFile), "The `to` parameter must be of type DatastoreFile"
		log.debug(f"<{self.path}> Copying to \"{to.path}\"...")
		if not self.exists:
			raise exceptions.DatastoreFileNotFoundError(self.datastore, self.relpath)
		
		file_manager = self.datastore._client._service_instance.RetrieveContent().fileManager
		self.datastore._client._wait_for_task(file_manager.CopyDatastoreFile_Task(
			sourceName=self.path,
			sourceDatacenter=self.datastore._datacenter,
			destinationName=to.path,
			destinationDatacenter=to.datastore._datacenter,
			force=force
		))

	def merge(self, to: 'DatastoreFile', force: bool = False):
		"""
		Merge (copy) the contents of this directory to another.

		:param to: The 'DatastoreFile' for the destination directory.
		:param force: If true, overwrite any identically named file at the destination.
		"""
		assert isinstance(to, DatastoreFile), "The `to` parameter must be of type DatastoreFile"
		log.debug(f"<{self.path}> Merging to \"{to.path}\"...")
		if not self.isdir:
			raise exceptions.DatastoreNotADirectoryError(self.datastore, self.relpath)
		if not to.isdir:
			raise exceptions.DatastoreNotADirectoryError(to.datastore, to.relpath)
		
		for file in self.files:
			file.copy(to / file.filename, force=force)

	def move(self, to: 'DatastoreFile', force: bool = False):
		"""
		Move this file or directory to another location.

		:param to: The 'DatastoreFile' for the destination.
		:param force: If true, overwrite any identically named file at the destination.
		"""
		assert isinstance(to, DatastoreFile), "The `to` parameter must be of type DatastoreFile"
		log.debug(f"<{self.path}> Moving to \"{to.path}\"...")
		if not self.exists:
			raise exceptions.DatastoreFileNotFoundError(self.datastore, self.relpath)
		
		file_manager = self.datastore._client._service_instance.RetrieveContent().fileManager
		self.datastore._client._wait_for_task(file_manager.MoveDatastoreFile_Task(
			sourceName=self.path,
			sourceDatacenter=self.datastore._datacenter,
			destinationName=to.path,
			destinationDatacenter=to.datastore._datacenter,
			force=force
		))

	def remove(self):
		"""
		Removes this file or directory on the datastore. Folder deletes are always recursive. If the path does not exist, this does nothing.
		"""
		log.debug(f"<{self.path}> Removing...")
		if not self.exists:
			return
		file_manager = self.datastore._client._service_instance.RetrieveContent().fileManager
		self.datastore._client._wait_for_task(file_manager.DeleteDatastoreFile_Task(
			name=self.path,
			datacenter=self.datastore._datacenter,
		))

	def register_vm(self, name: typing.Optional[str] = None) -> 'VirtualMachine':
		"""
		Register a new VM using this file. This must be a valid VMX file, or a directory containing a single VMX file

		:param name: The name to set for the VM. If `None`, the VM name will be derived from the VMX file.

		:return: A `VirtualMachine` object (or subtype) for the new VM.
		"""
		assert isinstance(name, str) or name is None

		log.debug(f"<{self.path}> Registering VM (name={name})...")
		if self.isfile and not self.filename.lower().endswith(".vmx"):
			raise exceptions.DatastoreError(self.datastore, f"{self.relpath} is not a valid .vmx file")
		
		target = self
		if self.isdir:
			files =  [ file for file in target.files if file.filename.lower().endwith(".vmx") ]
			if len(files) == 0:
				raise exceptions.DatastoreError(self.datastore, f"No .vmx file found in {target.relpath}")
			if len(files) > 1:
				raise exceptions.DatastoreError(self.datastore, f"Multiple .vmx files found in {target.relpath}: {', '.join([file.relpath for file in files])})")
			target = files[0]
			log.debug(f"<{self.path}> Registering VM from found VM file: {target.relpath}")
		
		destination_host = self.datastore._client._host_system
		vim_vm = self.datastore._client._wait_for_task(self.datastore._datacenter.vmFolder.RegisterVM_Task(
			path=target.path,
			name=name,
			asTemplate=False,
			pool=destination_host.parent.resourcePool,
			host=destination_host
		))
		return self.datastore._client.vms.get(str(vim_vm._moId), search_type='id')

	def __str__(self):
		return self.path

	def __repr__(self):
		return f"'{self.path}'"

	def _sanitize(self, path):
		return os.path.relpath(os.path.normpath(os.path.join("/", path)), "/").strip("/")

	@staticmethod
	def parse(client: 'ESXiClient', filepath: str) -> 'DatastoreFile':
		"""
		Convert a path to a `DatastoreFile`

		:param client: The `ESXiClient` to use.
		:param filepath: The file path to resolve. This may be in the form of an absolute path (e.g. `/vmfs/volumes/<id>/path/to/file`) or in the form of a datastore path (e.g. `[datastore] path/to/file`)

		:return: A new `DatastoreFile` object.
		"""
		if filepath.startswith("/"):
			# Absolute path
			# Try to detect the datastore
			PREFIX = "ds://"
			raw_filepath = filepath if filepath.startswith(PREFIX) else PREFIX + filepath
			for ds in client.datastores:
				ds_root_path = ds.path if ds.path.startswith(PREFIX) else PREFIX + ds.path
				if raw_filepath.startswith(ds_root_path):
					return DatastoreFile(ds, raw_filepath[len(ds_root_path):].lstrip("/"))
			raise ValueError(f"Unable to resolve datastore for absolute path {filepath}")
		
		if filepath.startswith("["):
			try:
				match = re.match("\[(" + "|".join([ re.escape(name) for name in client.datastores.names ]) + ")\]\s*(.+)", filepath, flags=re.IGNORECASE)
				datastore_name = match.group(1)
				relative_path = match.group(2).lstrip("/")
				return DatastoreFile(client.datastores.get(datastore_name), relative_path)
			except Exception as e:
				raise ValueError(f"Unable to resolve datastore path {filepath}: {str(e)}")
		raise ValueError(f"Unable to resolve path {filepath}")




from esxi_utils.util import exceptions
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.client import ESXiClient

class PhysicalNICList:
	"""
	The list of physical NICs on the ESXi host.
	"""
	def __init__(self, client: 'ESXiClient'):
		self._client = client

	def __iter__(self) -> typing.Iterator['PhysicalNIC']:
		return iter([ PhysicalNIC(self._client, pnic.device) for pnic in self._network_system.networkInfo.pnic ])

	def __getitem__(self, name) -> 'PhysicalNIC':
		return self.get(name)

	def __contains__(self, name: str) -> bool:
		return self.exists(name)

	@property
	def items(self) -> typing.List['PhysicalNIC']:
		"""
		A list of all items
		"""
		return list(self)

	def find(self, name: str) -> typing.Union['PhysicalNIC', None]:
		"""
		Get a Physical NIC by name.

		:param name: The name of the Physical NIC.

		:return: A `PhysicalNIC` object, or `None` if not found.
		"""
		found = [ pnic for pnic in self if pnic.name == name ]
		if len(found) != 1:
			return None
		return found[0]

	def get(self, name: str) -> 'PhysicalNIC':
		"""
		Get a Physical NIC by name and raise an exception if not found.

		:param name: The name of the Physical NIC.

		:return: A `PhysicalNIC` object
		"""
		pnic = self.find(name)
		if pnic is None:
			raise exceptions.NetworkingObjectNotFoundError('PhysicalNIC', name)
		return pnic

	def exists(self, name: str) -> bool:
		"""
		Check whether a Physical NIC exists 

		:parma name: The name of the Physical NIC

		:return: Whether or not the Physical NIC exists
		"""
		return self.find(name) is not None

	@property
	def _network_system(self):
		return self._client._host_system.configManager.networkSystem

	def __str__(self):
		return f"<{type(self).__name__} ({len(list(self))} Physical NICs)>"
	
	def __repr__(self):
		return str(self)


class PhysicalNIC:
	"""
	A physical network adapters as seen by the primary operating system. 
	"""
	def __init__(self, client: 'ESXiClient', name: str):
		self._client = client
		self._name = name

	@property
	def name(self) -> str:
		"""
		The device name of the physical network adapter.
		"""
		return self._name

	@property
	def up(self) -> bool:
		"""
		Whether or not this link is up.
		"""
		return (self._obj.linkSpeed is not None)

	@property
	def linkspeed(self) -> typing.Union[int, None]:
		"""
		Bit rate on the link, in megabits per second. If None, then the link is down. 
		"""
		if self._obj.linkSpeed:
			return self._obj.linkSpeed.speedMb
		return None

	@property
	def fullduplex(self) -> typing.Union[bool, None]:
		"""
		Flag to indicate whether or not the link is capable of full-duplex ("true") or only half-duplex ("false"). If None, then the link is down. 
		"""
		if self._obj.linkSpeed:
			return self._obj.linkSpeed.duplex
		return None
		
	@property
	def mac(self) -> str:
		"""
		The media access control (MAC) address of the physical network adapter. 
		"""
		return self._obj.mac

	@property
	def pci(self) -> str:
		"""
		Device hash of the PCI device corresponding to this physical network adapter. 
		"""
		return self._obj.pci

	@property
	def driver(self) -> str:
		"""
		The name of the driver.
		"""
		return self._obj.driver

	@property
	def _obj(self):
		for pnic in self._network_system.networkInfo.pnic:
			if pnic.device == self.name:
				return pnic
		raise exceptions.NetworkingObjectNotFoundError('PhysicalNIC', self.name)

	@property
	def _network_system(self):
		return self._client._host_system.configManager.networkSystem

	def __str__(self):
		return f"<{type(self).__name__} \"{self.name}\">"
	
	def __repr__(self):
		return str(self)


from esxi_utils.util.connect.ssh import SSHConnection
from dateutil.parser import parse as parsetime
import pkg_resources
import datetime
import textfsm
import typing
import re

if typing.TYPE_CHECKING:
	from esxi_utils.util.response import Response

class CiscoSSHConnection(SSHConnection):
	"""
	A SSH connection to a Cisco remote host.

	:param ip: The IP of the target.
	:param username: The username to use when logging into the remote system.
	:param password: The password to use when logging into the remote system.
	:param port: The port to use when connecting over SSH.
	"""

	def exec(
		self, 
		cmd: str, 
		stdin: typing.Optional[typing.Dict[str, str]] = None, 
		timeout: int = 120, 
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None, 
		remove_banner: bool = True,
		network_retries: int = 0,
	) -> 'Response':
		"""
		Execute a shell command on the remote end of this connection. Includes some Cisco-specific functionality, such as ignoring login headers.

		:param cmd: The command to run.
		:param stdin: A dict mapping a 'pattern' (regex string) to a 'response' (string), to input 'response' to stdin when stdout matches 'pattern'
		:param timeout: The command timeout in seconds. Set to 0 to disable timeout.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available.
		:param remove_banner: Whether or not to auto-detect and remove login banners from the resulting output.
		:param network_retries: Number of times to retry the command when a recoverable network error is encountered. The command should be idempotent when this is non-zero as it may run several times.

		:return: A `Response` object.
		"""
		banner = None
		if remove_banner:
			banner_resp = SSHConnection.exec(self, cmd="show clock", network_retries=3).stdout.replace('\r', '').rstrip()
			banner_end = banner_resp.rfind("\n") # Find the last newline
			if banner_end != -1:
				# The banner is everything up to and including the last newline character, 
				# since the 'show clock' command predictably outputs a single line without any control characters
				banner = banner_resp[:banner_end+1]

		resp = SSHConnection.exec(self, cmd=cmd, stdin=stdin, timeout=timeout, pty=False, out_stream_callback=out_stream_callback, network_retries=network_retries)
		resp.stdout = resp.stdout.replace('\r', '')
		if banner:
			resp.stdout = resp.stdout[len(banner):]
		return resp

	def show_ip_interface_brief(self) -> typing.List[typing.Dict[str, typing.Any]]:
		"""
		Finds the ip interface information
		
		:return: A list of objects containing interface information with the following fields:

		- Interface

		- IP-Address

		- OK?

		- Method

		- Status
		
		- Protocol

		"""
		return self._parse_table(
			stringtable=self.exec("show ip interface brief", network_retries=3).stdout, 
			headers=['Interface', 'IP-Address', 'OK?', 'Method', 'Status', 'Protocol']
		)

	def show_ip_ospf_neighbor(self) -> typing.List[typing.Dict[str, typing.Any]]:
		"""
		Finds the ospf neighbors
		
		:return: A list of objects containing interface information with the following fields:

		- Address

		- Dead Time

		- Interface

		- Neighbor ID

		- Pri

		- State

		"""
		return self._parse_table(
			stringtable=self.exec("show ip ospf neighbor", network_retries=3).stdout, 
			headers=['Neighbor ID', 'Pri', 'State', 'Dead Time', 'Address', 'Interface']
		)

	def show_license_usage(self) -> typing.Dict[str, typing.Any]:
		"""
		Finds the license usage.
		
		:return: The license usage for the cisco machine.
		"""
		parsed_data = self._parse_set(
			set_string=self.exec("show license usage", network_retries=3).stdout, 
			first_line='License Authorization:'
		)
		return_obj = dict()
		return_obj['Licenses'] = [ value for field, value in parsed_data.items() if field != 'License Authorization']
		return return_obj

	def get_logging_trap_info(self) -> typing.List[typing.Dict[str, typing.Any]]:
		"""
		Returns the logging trap info from show logging
		
		:return: Logging trap info summary
		"""
		output = self.exec("show logging", network_retries=3).stdout
		template = pkg_resources.resource_filename(__name__, 'textfsm_templates/show_logging_trap_info.textfsm')
		with open(template) as f:
			re_table = textfsm.TextFSM(f)
			header = re_table.header
			results = re_table.ParseText(output)
			return [{ header[i]: r[i] for i in range(len(header)) } for r in results[:-1] ]

	def get_flow_exporter_info(self) -> typing.List[typing.Dict[str, typing.Any]]:
		"""
		Returns the flow exporter info

		:return: Flow exporter info summary
		"""
		output = self.exec("show run flow exporter", network_retries=3).stdout
		template = pkg_resources.resource_filename(__name__, 'textfsm_templates/show_run_flow_exporter.textfsm')
		with open(template) as f:
			re_table = textfsm.TextFSM(f)
			header = re_table.header
			results = re_table.ParseText(output)
			return [{ header[i]: r[i] for i in range(len(header)) } for r in results ]

	def time(self):
		"""
		Get the current time set on the remote host.

		:return: A datetime object
		"""
		date = self.exec("show clock", network_retries=3).stdout.splitlines()[-1].strip()
		if date.startswith('*'):
			date = date[1:]
		return parsetime(date).astimezone(datetime.timezone.utc)

	def _parse_table(self, stringtable, headers):
		"""
		Parses a table. This assumes data is contained within the headers, e.g.
		header1          header2
		value1           value2
		superlongvalue1  superlongvalue2

		object2:
		 subfield3:subvalue3

		:param stringtable: String that contains table to be parsed
		:param headers: List of headers in tables
		
		:return: A list of objects whose fields match the headers.
		"""
		# Remove empty lines and whitespace lines
		lines = list(filter(lambda str: str and not str.isspace(),stringtable.splitlines()))
		regex = re.compile('\s*'+'\s*'.join([re.escape(head) for head in headers])+'\s*')
		
		# Find the header of the table
		while lines and not regex.match(lines[0]):
			lines.pop(0)

		if len(lines)==0:
			raise ValueError(f"Table with headers {headers} not found in string {stringtable}")

		# Find where each header starts
		head_index = [lines[0].find(head) for head in headers]
		# Add the end of the line
		head_index.append(None)
		
		return [ { headers[idx]: line[head_index[idx]:head_index[idx+1]].strip() for idx in range(len(headers)) } for line in lines[1:] ]

	def _parse_set(self, set_string, indent=2, first_line=None):
		"""
		Parses a set, e.g. data in the form
		object1:
		  subfield1:subvalue1
		  subfield2:subvalue2
		  subobject1:
		    subsubfield1:subsubvalue1

		object2:
		  subfield3:subvalue3

		:set_string: String to parse
		:indent: Default 2. Number of characters per left indent.
		
		:return:
			The set converted to a dictionary. In the above example, the result would be::

				{
					object1:
					{
						subfield1:subvalue1,
						subfield2:subvalue2,
						subobject1:{
							subsubfield1:subsubvalue1
						}

					},
					object2:{
						subfield3:subvalue3
					}
				}
		"""
		lines = list(filter(lambda str: str and not str.isspace(),set_string.splitlines()))

		# Find where the set begins
		while lines and first_line and lines[0].strip()!=first_line:
			lines.pop(0)

		# Start with an empty object
		obj_stack = [{}]

		def _num_white_space(line):
			return len(line)-len(line.lstrip())

		for line in lines:
			split = line.split(':')

			if len(split)==1:
				raise ValueError(f'Each line of set_string must be of the form field:value or field:\n{set_string}')

			field = split[0].strip()
			value = split[1].strip()

			if _num_white_space(line)%indent !=0:
				raise ValueError(f'Error with line \n{line}\n. Each line must have a left indent of multiple {indent}. This line has an indent with {_num_white_space(line)} characters.')
			# The length of the stack (minus 1) matches the depth of the current
			# object. If the indent is less, pop the obj_stack until
			# you get the object that matches the current depth. 
			while _num_white_space(line) < (len(obj_stack)-1)*indent:
				obj_stack.pop()

			# If there is no value, that means we have a new object.
			if not value:
				obj_stack[-1][field] = {}
				obj_stack.append(obj_stack[-1][field])
			# Otherwise we add the current field/value to the current object
			else:
				obj_stack[-1][field] = value
		
		# Return the root object
		return obj_stack[0]


from esxi_utils.vm.tools.guesttools import GuestTools, GuestToolsResponsePromise
from esxi_utils.util.response import Response
from esxi_utils.util import log, exceptions
import typing
import stat
import os
import re

class UnixGuestTools(GuestTools):
	"""
	Guest tools subclass specifically for interaction with a Unix or Unix-like operating system

	Wrapper class for functionality related to VMware Tools on a Virtual Machine. The virtual machine must have VMware tools installed for
	any of the contained functions to work.

	:param vm: A `VirtualMachine` object to wrap. VMware tools must be installed on this virtual machine.
	"""
	def bash_async(
		self, 
		username: str,
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None,
		bash_path: str = "/bin/bash",
		sudo: typing.Union[bool, str] = False,
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	) -> 'GuestToolsBashResponsePromise':
		"""
		Runs a script in the guest operating system asynchronously using bash.

		Unlike ``execute_program``, the command is executed as a standard bash script and thus the full path to programs do not need to be provided. 
		Systems without bash will fail. ``command`` may also be provided as a script, with newlines separating the commands to be run.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param command: The command or script to run.
		:param cwd: The working directory that the command should be run in.
		:param bash_path: The path to the bash executable.
		:param sudo: Whether to run the command with sudo. If `True`, the `password` parameter will be passed to sudo. Provide a string to use a different password for sudo.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
	
		:return: A `GuestToolsBashResponsePromise` object
		"""
		return GuestToolsBashResponsePromise(
			tools=self, 
			username=username, 
			password=password, 
			command=command, 
			cwd=cwd,
			bash_path=bash_path, 
			sudo=sudo,
			out_stream_callback=out_stream_callback
		)

	def bash(
		self, 
		username: str,
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None,
		bash_path: str = "/bin/bash",
		sudo: typing.Union[bool, str] = False,
		timeout: typing.Optional[int] = 120,
		assert_status: typing.Optional[int] = None,
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	) -> 'Response':
		"""
		Runs a script in the guest operating system using bash. Unlike ``bash_async``, this will block until the associated command has finished.

		Unlike ``execute_program``, the command is executed as a standard bash script and thus the full path to programs do not need to be provided. 
		Systems without bash will fail. ``command`` may also be provided as a script, with newlines separating the commands to be run.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param command: The command or script to run.
		:param cwd: The working directory that the command should be run in.
		:param bash_path: The path to the bash executable.
		:param sudo: Whether to run the command with sudo. If `True`, the `password` parameter will be passed to sudo. Provide a string to use a different password for sudo.
		:param timeout: The command timeout in seconds. Set to 0 or `None` to disable timeout.
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 

		:return: A `Response` object
		"""
		return self.bash_async(
			username=username, 
			password=password, 
			command=command, 
			cwd=cwd,
			bash_path=bash_path, 
			sudo=sudo,
			out_stream_callback=out_stream_callback
		).wait(assert_status=assert_status, timeout=timeout)

	def stat(self, path: str, username: str, password: str, bash_path: str = "/bin/bash", sudo: typing.Union[bool, str] = False) -> typing.Union[typing.Dict[str, typing.Any], None]:
		"""
		Get information about a file.

		:param path: A path to a file or directory on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param bash_path: The path to the bash executable.
		:param sudo: Whether to run commands with sudo. If `True`, the `password` parameter will be passed to sudo. Provide a string to use a different password for sudo.

		:return: A dict containing information about the file, or `None` if the file does not exist.
		"""
		log.debug(f"{str(self)} Stat {path}")
		response = self.bash(
			username=username,
			password=password,
			command=f"stat -c '%s %f %Y %X %g %u' '{path}'",
			bash_path=bash_path,
			sudo=sudo
		)
		output = str(response.stdout) + "\n" + str(response.stderr)
		output = output.strip()
		if response.status != 0 and f"no such file" in output.lower():
			return None
		elif response.status != 0:
			raise exceptions.RemoteConnectionCommandError(self, response)
		
		split = output.split(" ")
		size = int(split[0])
		mode = int(split[1], 16)
		mtime = int(split[2])
		atime = int(split[3])
		gid = int(split[4])
		uid = int(split[5])

		return {
			"name": os.path.basename(path),
			"size": size,
			"mode": mode,
			"isfile": stat.S_ISREG(mode),
			"isdir": stat.S_ISDIR(mode),
			"mtime": mtime,
			"atime": atime,
			"gid": gid,
			"uid": uid
		}

	def bulkstat(self, paths: typing.List[str], username: str, password: str, bash_path: str = "/bin/bash", sudo: typing.Union[bool, str] = False) -> typing.List[typing.Dict[str, typing.Any]]:
		"""
		Get information about one or more files.

		:param paths: The paths to the files or directories on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param bash_path: The path to the bash executable.
		:param sudo: Whether to run commands with sudo. If `True`, the `password` parameter will be passed to sudo. Provide a string to use a different password for sudo.

		:return: A dict containing information about the file, or `None` if the file does not exist.
		"""
		paths_string = " ".join([ f"'{path}'" for path in paths ])
		log.debug(f"{str(self)} Bulk Stat: {', '.join(paths)}")

		with self.use_temporary_file(username=username, password=password, extension=".stat") as output_file:
			self.bash(
				username=username,
				password=password,
				command=f"stat -c '%s %f %Y %X %g %u' {paths_string} > {output_file} 2>&1",
				bash_path=bash_path,
				sudo=sudo
			)

			output: str = self.get_file(username=username, password=password, filepath=output_file) # type: ignore

		lines = output.strip().split("\n")
		if len(lines) != len(paths):
			raise RuntimeError(f"bulkstat: the number of output lines do not match the number of input files")

		stats = list()
		for path, line in zip(paths, lines):
			obj = {
				"path": path,
				"name": os.path.basename(path),
				"exists": True,
				"size": None,
				"mode": None,
				"isfile": None,
				"isdir": None,
				"mtime": None,
				"atime": None,
				"gid": None,
				"uid": None
			}

			if f"no such file" in line.lower():
				obj["exists"] = False
				stats.append(obj)
				continue

			try:
				split = line.split(" ")
				obj["size"] = int(split[0])
				obj["mode"] = int(split[1], 16)
				obj["isfile"] = stat.S_ISREG(obj["mode"])
				obj["isdir"] = stat.S_ISDIR(obj["mode"])
				obj["mtime"] = int(split[2])
				obj["atime"] = int(split[3])
				obj["gid"] = int(split[4])
				obj["uid"] = int(split[5])
				stats.append(obj)
			except Exception:
				raise RuntimeError(f"Failed to stat file {path}: {line}")
		return stats
	
	def isfile(self, path: str, username: str, password: str, bash_path: str = "/bin/bash", sudo: typing.Union[bool, str] = False) -> bool:
		"""
		Checks if the file exists and is a regular file.

		:param path: A path to a file on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param bash_path: The path to the bash executable.
		:param sudo: Whether to run commands with sudo. If `True`, the `password` parameter will be passed to sudo. Provide a string to use a different password for sudo.
		
		:return: A boolean whether or not the file exists.
		"""
		stat = self.stat(
			path, 
			username=username, 
			password=password,
			bash_path=bash_path,
			sudo=sudo
		)
		if stat is None:
			return False
		return stat["isfile"]

	def isdir(self, path: str, username: str, password: str, bash_path: str = "/bin/bash", sudo: typing.Union[bool, str] = False) -> bool:
		"""
		Checks if the directory exists.

		:param path: A path to a directory on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param bash_path: The path to the bash executable.
		:param sudo: Whether to run commands with sudo. If `True`, the `password` parameter will be passed to sudo. Provide a string to use a different password for sudo.
		
		:return: A boolean whether or not the directory exists.
		"""
		stat = self.stat(
			path, 
			username=username, 
			password=password,
			bash_path=bash_path,
			sudo=sudo
		)
		if stat is None:
			return False
		return stat["isdir"]

	def download(
		self, 
		path: str, 
		dst: str, 
		username: str,
		password: str,
		directory_contents_only: bool = False, 
		overwrite: bool = False,
		bash_path: str = "/bin/bash",
	) -> typing.List[str]:
		"""
		Download a file or directory from the remote system.

		:param path: Path to the file or directory on the remote system.
		:param dst: The local destination file or directory to download to.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param directory_contents_only: If `True` and `path` points to a directory, then just the contents of the directory will be downloaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files.
		:param bash_path: The path to the bash executable.

		:return: A list of local paths to all files downloaded.
		"""
		log.info(f"{str(self)} Downloading \"{path}\" to \"{dst}\"...")
		dst = os.path.abspath(dst)
		basename = os.path.basename(path)
		stat = self.stat(path, username=username, password=password, bash_path=bash_path)
		if stat is None:
			raise exceptions.RemoteFileNotFoundError(self, path)

		download_map: typing.Dict[str, str] = {} # Map download source -> destination
		if stat["isfile"]:
			# Single file download
			target = dst
			if not os.path.exists(target):
				# Target doesn't exist, we will download under the name specified
				# Just ensure the parent directory does exist
				parent = os.path.dirname(target)
				if not os.path.isdir(parent):
					raise NotADirectoryError(f"{parent} is not a valid local directory")
			else:
				# Target exists, we'll download into the target with the same name
				if not os.path.isdir(target):
					raise NotADirectoryError(f"{target} is not a valid local directory")
				target = os.path.join(target, basename)
			download_map[path] = target
		else:
			# Directory download
			target = dst
			if os.path.isdir(target) and directory_contents_only:
				# Downloading directory contents into a directory
				pass
			elif os.path.isdir(target) and not directory_contents_only:
				# Downloading directory into a directory, download with the same name into the target
				target = os.path.join(target, stat["name"])
			elif not os.path.exists(target) and directory_contents_only:
				# We cannot download the contents of a directory if the dst does not exist
				raise NotADirectoryError(f"{dst} is not a valid local directory")
			elif not os.path.exists(target) and not directory_contents_only:
				# Downloading directory under a new name, we just need to ensure that the parent directory exists
				parent = os.path.dirname(target)
				if not os.path.isdir(parent):
					raise NotADirectoryError(f"{parent} is not a valid local directory")
			
			download_map = {}
			files_to_download = [ os.path.join(path, filename) for filename in self.list_files(username=username, password=password, filepath=path) ]
			while len(files_to_download):
				stats = self.bulkstat(files_to_download, username=username, password=password, bash_path=bash_path)
				files_to_download = []
				for stat in stats:
					if stat["isdir"]:
						files_to_download.extend([ os.path.join(stat["path"], filename) for filename in self.list_files(username=username, password=password, filepath=stat["path"]) ])
						continue
					download_map[stat["path"]] = os.path.join(target, stat["path"][len(path):].strip("/"))

		if len(download_map.keys()) == 0:
			return []

		for target in download_map.values():
			if os.path.exists(target) and not overwrite:
				raise FileExistsError(f"{target} already exists")
			if os.path.isdir(target):
				raise IsADirectoryError(f"{target} cannot be overwritten as it exists as a directory")

		# Begin downloading files
		for remotepath, localpath in download_map.items():
			log.debug(f"{self} Downloading {remotepath} to {localpath}")
			os.makedirs(os.path.dirname(localpath), exist_ok=True)
			file_bytes: bytes = self.get_file(username=username, password=password, filepath=remotepath, encoding=None) # type: ignore
			with open(localpath, mode="wb") as f:
				f.write(file_bytes)

		return list(download_map.values())

	def upload(
		self, 
		src: str,
		dst: str,
		username: str,
		password: str,
		directory_contents_only: bool = False, 
		overwrite: bool = False,
		bash_path: str = "/bin/bash",
	) -> typing.List[str]:
		"""
		Upload a local file or directory to this path on the remote system.
		
		Warning: This function will load entire files into memory. Ensure that extremely large files are not uploaded using this method.

		:param src: The local path to a file or directory to upload.
		:param dst: The remote path to upload to.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param directory_contents_only: If `True` and `src` points to a directory, then only the contents of the directory will be uploaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files on the remote system.
		:param bash_path: The path to the bash executable.

		:return: A list of remote paths to all files uploaded.
		"""
		log.info(f"<{self}> Uploading \"{src}\" to \"{dst}\"...")
		src = os.path.abspath(src)
		if not os.path.exists(src):
			raise FileNotFoundError(f"File or directory {src} not found")

		dst_stat = self.stat(dst, username=username, password=password, bash_path=bash_path)

		upload_map: typing.Dict[str, str] = {} # Map upload source -> destination
		if os.path.isfile(src):
			# Single file upload
			target = dst
			if dst_stat is None:
				# Target doesn't exist, we will upload under the name specified
				# Just ensure the parent directory does exist
				parent = os.path.dirname(target)
				if not self.isdir(parent, username=username, password=password, bash_path=bash_path):
					raise exceptions.RemoteNotADirectoryError(self, parent)
			else:
				# Target exists, we'll upload into the target with the same name
				if not dst_stat["isdir"]:
					raise exceptions.RemoteNotADirectoryError(self, dst)
				target = os.path.join(target, os.path.basename(src))
			upload_map[src] = target
		elif os.path.isdir(src):
			# Directory upload
			target = dst
			target_exists = dst_stat is not None
			target_is_dir = dst_stat and dst_stat["isdir"]
			if target_is_dir and directory_contents_only:
				# Uploading directory contents into a directory
				pass
			elif target_is_dir and not directory_contents_only:
				# Uploading directory into a directory, upload with the same name into the target
				target = os.path.join(target, os.path.basename(src))
			elif not target_exists and directory_contents_only:
				# We cannot upload the contents of a directory if the dst does not exist
				raise exceptions.RemoteNotADirectoryError(self, target)
			elif not target_exists and not directory_contents_only:
				# Uploading directory under a new name, we just need to ensure that the parent directory exists
				parent = os.path.dirname(target)
				if not self.isdir(parent, username=username, password=password, bash_path=bash_path):
					raise exceptions.RemoteNotADirectoryError(self, parent)
			
			for root, _, files in os.walk(src):
				for file in files:
					path = os.path.join(root, file)
					upload_map[path] = os.path.join(target, path[len(src):].strip("/"))

		if len(upload_map.keys()) == 0:
			return []

		for stat in self.bulkstat(list(upload_map.values()), username=username, password=password):
			if stat["exists"] and stat["isfile"] and not overwrite:
				raise exceptions.RemoteFileExistsError(self, stat["path"])
			if stat["exists"] and stat["isdir"]:
				raise exceptions.RemoteIsADirectoryError(self, stat["path"])

		# Determine the directories to create
		# We will be running mkdir with parents, so we only want the longest paths
		dirs_to_make = []
		for remotepath in upload_map.values():
			dirname = os.path.dirname(remotepath).rstrip("/") + "/"
			dirs_to_make = [ path for path in dirs_to_make if not dirname.startswith(path) ] # Remove any paths that are subsets of this path 
			is_subpath = any([ path.startswith(dirname) for path in dirs_to_make ]) # Check if this is a subpath of any existing path already
			if not is_subpath:
				dirs_to_make.append(dirname)
		
		# Make directories
		mkdir_path_string = " ".join([ f"'{path}'" for path in dirs_to_make ])
		self.bash(
			username=username, 
			password=password,
			command=f"mkdir -p {mkdir_path_string}",
			bash_path=bash_path,
			assert_status=0
		)

		# Begin uploading files
		for localpath, remotepath in upload_map.items():
			log.debug(f"<{self}> Uploading \"{localpath}\" to \"{remotepath}\"")
			with open(localpath, "rb") as f:
				self.write_file(username=username, password=password, filepath=remotepath, data=f.read())

		return list(upload_map.values())
	

class GuestToolsBashResponsePromise(GuestToolsResponsePromise):
	"""
	Encapsulates a Promise for a result of an GuestTools Bash command.

	:param tools: The UnixGuestTools instance from which this promise originates.
	:param username: The username of the virtual machine user.
	:param password: The password of the virtual machine user.
	:param command: The command or script to run.
	:param bash_path: The path to the bash executable.
	:param sudo: Whether to run the command with sudo. If `True`, the `password` parameter will be passed to sudo. Provide a string to use a different password for sudo.
	:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
	"""
	def __init__(
		self, 
		tools: UnixGuestTools, 
		username: str, 
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None, 
		bash_path: str = "/bin/bash", 
		sudo: typing.Union[bool, str] = False,
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	):
		assert isinstance(cwd, str) or cwd is None, "cwd must be a string or None"
		assert isinstance(bash_path, str), "bash_path must be a string"
		assert isinstance(sudo, (str, bool)), "sudo must be a string or boolean"

		self._bash_path = bash_path
		self._sudo = sudo
		self._scriptfile = None
		
		super().__init__(
			tools=tools, 
			username=username, 
			password=password, 
			command=command,
			cwd=None,
			make_output_available=True,
			out_stream_callback=out_stream_callback
		)

	def _setup(self, make_output_available: bool = False) -> str:
		"""
		Perform any setup necessary prior to starting the command.

		:param make_output_available: Whether or not it is requested to make the output of the command available.
		
		:return: The command to run.
		"""
		self._scriptfile = self._tools.create_temporary_file(self._username, self._password, ".sh")
		self._tools.write_file(username=self._username, password=self._password, filepath=self._scriptfile, data=f"cd {self._cwd}\n{self._command}" if self._cwd else self._command)

		command = f"{self._bash_path} -eu {self._scriptfile}"

		# Handle sudo
		if self._sudo:
			# As root; invoke the script with sudo
			sudo_password = self._sudo if isinstance(self._sudo, str) else self._password
			command = f"{self._bash_path} -c \"echo '{sudo_password}' | sudo -S {command}\""

		# Handle stdout/stderr
		if make_output_available:
			self._stdoutfile = self._tools.create_temporary_file(self._username, self._password, ".out")
			self._stderrfile = self._tools.create_temporary_file(self._username, self._password, ".err")
			command = f"{command} > {self._stdoutfile} 2> {self._stderrfile}"

		return command

	def _format_response(self, response: Response) -> Response:
		"""
		Perform post-processing on a result. This will occur when a command finishes executing but before any files are cleaned up.

		:param response: The result of the command.

		:return: The new result.
		"""
		if self._sudo:
			response.stdout = re.sub(r"^\[sudo\] password[^\n]+\n?", "", str(response.stdout))
			response.stderr = re.sub(r"^\[sudo\] password[^\n]+\n?", "", str(response.stderr))
		return response
	
	def _cleanup(self):
		"""
		Clean up any files or other resources after the process has finished.
		"""
		super()._cleanup()
		if self._scriptfile:
			self._tools.delete_file(self._username, self._password, self._scriptfile)
			self._scriptfile = None



from esxi_utils.firewall.firewall import Firewall
from esxi_utils.firewall.ruleset import RulesetList, Ruleset
from esxi_utils.firewall.rule import RuleList, Rule

__all__ = [
	"Firewall",
	"RulesetList",
	"Ruleset",
	"RuleList",
	"Rule"
]


from esxi_utils.vm.hardware.cdrom import VirtualCdrom, VirtualCdromList
from esxi_utils.vm.hardware.device import VirtualDevice, VirtualDeviceList
from esxi_utils.vm.hardware.disk import VirtualDisk, VirtualDiskList
from esxi_utils.vm.hardware.floppy import VirtualFloppy, VirtualFloppyList
from esxi_utils.vm.hardware.nic import VirtualNIC, VirtualNICList
from esxi_utils.vm.hardware.video_card import VirtualVideoCard, VirtualVideoCardList

__all__ = [
	"VirtualCdrom",
	"VirtualDevice",
	"VirtualDisk",
	"VirtualFloppy",
	"VirtualNIC",
    "VirtualVideoCard"
]